"""Code for training and evaluating networks."""
import logging
import os
import flammkuchen
import numpy as np
from . import utils, data, models, event_utils, segment_utils, annot
from typing import List, Optional, Dict, Any, Sequence, Iterable, Union
import glob
import tensorflow
import librosa
import zarr
from tqdm.autonotebook import tqdm
import dask.config
import dask.array as da
from dask.diagnostics import ProgressBar

dask.config.set(**{"array.slicing.split_large_chunks": True})


def predict_probabilities(
    x: np.ndarray,
    model: tensorflow.keras.Model,
    params: Dict[str, Any],
    verbose: Optional[int] = 1,
    prepend_data_padding: bool = True,
):
    """[summary]

    Args:
        x ([samples, ...]): [description]
        model (tf.keras.Model): [description]
        params ([type]): [description]
        verbose (int, optional): Verbose level for predict() (see keras docs). Defaults to 1.
        prepend_data_padding (bool, optional): Restores samples that are ignored
                    in the beginning of the first and the end of the last chunk
                    because of "ignore_boundaries". Defaults to True.
    Returns:
        y_pred - output of network for each sample [samples, nb_classes]
    """

    pred_gen = data.AudioSequence(x=x, y=None, shuffle=False, **params)  # prep data
    nb_batches = len(pred_gen)
    verbose = 1
    prepend_data_padding = True

    temp_store = zarr.storage.TempStore()
    class_probabilities = zarr.zeros(
        shape=(0, params["nb_classes"]), chunks=(100_000, params["nb_classes"]), dtype="float", store=temp_store, overwrite=True
    )
    # predict and unpack batch by batch to memmaped np or zarr array
    for batch_number, batch_data in tqdm(enumerate(pred_gen), total=nb_batches, disable=verbose < 1):
        y_pred_batch = model.predict_on_batch(batch_data)  # run the network

        # reshape from [batches, nb_hist, ...] to [time, ...]
        y_pred_unpacked_batch = data.unpack_batches(y_pred_batch, pred_gen.data_padding)

        if prepend_data_padding:
            pad_width = None
            if batch_number == 0:  # to account for loss of samples at the first and last chunks
                pad_width = ((params["data_padding"], 0), (0, 0))
            elif batch_number == nb_batches - 1:
                pad_width = ((0, params["data_padding"]), (0, 0))

            if pad_width is not None:
                y_pred_unpacked_batch = np.pad(y_pred_unpacked_batch, pad_width=pad_width, mode="constant", constant_values=0)

        class_probabilities.append(y_pred_unpacked_batch, axis=0)

    class_probabilities = da.from_zarr(class_probabilities, inline_array=True)
    class_probabilities = class_probabilities.rechunk((1_000_000, params["nb_classes"]))
    return class_probabilities


def labels_from_probabilities(
    probabilities, threshold: Optional[float] = None, indices: Optional[Union[Sequence[int], slice]] = None
) -> np.ndarray:
    """Convert class-wise probabilities into labels.

    Args:
        probabilities ([type]): [samples, classes] or [samples, ]
        threshold (float, Optional): Argmax over all classes (Default, 2D - corresponds to 1/nb_classes or 0.5 if 1D).
                                     If float, each class probability is compared to the threshold.
                                     First class to cross threshold wins.
                                     If no class crosses threshold label will default to the first class.
        indices: (List[int], Optional): List of indices into axis 1 for which to compute the labels.
                                     Defaults to None (use all indices).
    Returns:
        labels [samples,] - index of "winning" dimension for each sample
    """
    if indices is None:
        indices = slice(None)  # equivalent to ":"

    if probabilities.ndim == 1:
        if threshold is None:
            threshold = 0.5
        labels = (probabilities[:, indices] > threshold).astype(np.intp)
    elif probabilities.ndim == 2:
        if threshold is None:
            labels = np.argmax(probabilities[:, indices], axis=1)
        else:
            thresholded_probabilities = probabilities[:, indices].copy()
            thresholded_probabilities[thresholded_probabilities < threshold] = 0
            labels = np.argmax(thresholded_probabilities > threshold, axis=1)
    else:
        raise ValueError(f"Probabilities have to many dimensions ({probabilities.ndim}). Can only be 1D or 2D.")

    return labels


def predict_segments(
    class_probabilities: np.ndarray,
    samplerate: float = 1.0,
    segment_dims: Optional[Sequence[int]] = None,
    segment_names: Optional[Sequence[str]] = None,
    segment_ref_onsets: Optional[List[float]] = None,
    segment_ref_offsets: Optional[List[float]] = None,
    segment_thres: float = 0.5,
    segment_minlen: Optional[float] = None,
    segment_fillgap: Optional[float] = None,
    segment_labels_by_majority: bool = True,
) -> Dict:
    """[summary]

    TODO: document different approaches for single-type vs. multi-type segment detection

    Args:
        class_probabilities ([type]): [T, nb_classes] with probabilities for each class and sample
                                      or [T,] with integer entries as class labels
        samplerate (float, optional): Hz. Defaults to 1.0.
        segment_dims (Optional[List[int]], optional): set of indices into class_probabilities corresponding
                                                      to segment-like song types.
                                                      Needs to include the noise dim.
                                                      Required to ignore event-like song types.
                                                      Defaults to None (all classes are considered segment-like).
        segment_names (Optional[List[str]], optional): Names for segment-like classes.
                                                       Defaults to None (use indices of segment-like classes).
        segment_ref_onsets (Optional[List[float]], optional):
                            Syllable onsets (in seconds) to use for estimating labels.
                            Defaults to None (will use onsets est from class_probabilitieslabels as ref).
        segment_ref_offsets (Optional[List[float]], optional): [description].
                            Syllable offsets (in seconds) to use for estimating labels.
                            Defaults to None (will use offsets est from class_probabilitieslabels as ref).
        segment_thres (float, optional): [description]. Defaults to 0.5.
        segment_minlen (Optional[float], optional): seconds. Defaults to None.
        segment_fillgap (Optional[float], optional): seconds. Defaults to None.
        segment_labels_by_majority (bool, optional): Segment labels given by majority of label values within on- and offsets.
                                                     Defaults to True.

    Returns:
        dict['segmentnames']['denselabels-samples'/'onsets'/'offsets'/'probabilities']
    """
    probs_are_labels = class_probabilities.ndim == 1
    if segment_dims is None:
        if not probs_are_labels:  # class_probabilities is [T, nb_classes]
            nb_classes = class_probabilities.shape[1]
        else:  # class_probabilities is [T,] with integer entries as class labels
            nb_classes = int(da.max(class_probabilities).compute())
        segment_dims = list(range(nb_classes))

    if segment_names is None:
        segment_names = [str(sd) for sd in segment_dims]

    segments: Dict[str, Any] = dict()
    if len(segment_dims):
        segments["samplerate_Hz"] = samplerate
        segments["index"] = segment_dims
        segments["names"] = segment_names
        if not probs_are_labels:
            segments["probabilities"] = class_probabilities[:, segment_dims]
            labels = da.map_blocks(
                labels_from_probabilities, class_probabilities, segment_thres, segment_dims, dtype=np.intp, drop_axis=1
            )
        else:
            segments["probabilities"] = None
            labels = class_probabilities

        # turn into song (0), no song (1) sequence to detect onsets (0->1) and offsets (1->0)
        song_binary = (labels > 0).astype(np.int8)
        if segment_fillgap is not None:
            song_binary = da.map_overlap(
                segment_utils.fill_gaps,
                song_binary,
                gap_dur=int(segment_fillgap * samplerate),
                depth=(int(segment_fillgap * samplerate + 1), 0),
                boundary="none",
                trim=True,
                align_arrays=True,
            )
        if segment_minlen is not None:
            song_binary = da.map_overlap(
                segment_utils.remove_short,
                song_binary,
                min_len=int(segment_minlen * samplerate),
                depth=(int(segment_minlen * samplerate + 1), 0),
                boundary="none",
                trim=True,
                align_arrays=True,
            )

        # detect syllable on- and offsets
        # pre- and post-pend 0 so we detect on and offsets at boundaries
        onsets = da.where(da.diff(song_binary, prepend=0) == 1)[0]
        offsets = da.where(da.diff(song_binary, append=0) == -1)[0]
        logging.info("   Detecting syllable on and offsets:")
        with ProgressBar(minimum=5):
            onsets, offsets = da.compute(onsets, offsets)
        segments["onsets_seconds"] = onsets.astype(float) / samplerate
        segments["offsets_seconds"] = offsets.astype(float) / samplerate

        # there is just a single segment type plus noise - in that case we use the gap-filled, short-deleted pred
        sequence: List[int] = []  # default to empty list
        if len(segment_dims) == 2:
            labels = song_binary
            # syllable-type for each syllable as int
            sequence = [str(segment_names[1])] * len(segments["offsets_seconds"])
        # if >1 segment type (plus noise) label sylls by majority vote on un-smoothed labels
        elif len(segment_dims) > 2 and segment_labels_by_majority:
            # if no refs provided, use use on/offsets from smoothed labels
            if segment_ref_onsets is None:
                segment_ref_onsets = segments["onsets_seconds"]
            if segment_ref_offsets is None:
                segment_ref_offsets = segments["offsets_seconds"]

            logging.info("   Labeling by majority:")
            if len(segment_dims) < np.iinfo("uint8").max:
                cast_to = np.uint8
            elif len(segment_dims) < np.iinfo("uint16").max:
                cast_to = np.uint16
            elif len(segment_dims) < np.iinfo("uint32").max:
                cast_to = np.uint32
            else:
                cast_to = None

            if cast_to is not None:
                logging.info(f"   Casting labels to {cast_to}:")
                labels = da.map_blocks(lambda x: x.astype(cast_to), labels, dtype=np.int16)

            with ProgressBar(minimum=5):
                labels = labels.compute()

            # syllable-type for each syllable as int
            sequence, labels = segment_utils.label_syllables_by_majority(
                labels, segment_ref_onsets, segment_ref_offsets, samplerate
            )
        segments["samples"] = labels
        segments["sequence"] = sequence
    return segments


def _detect_events_oom(
    event_probability: np.ndarray, thres: float = 0.70, min_dist: int = 100, index: int = 0, block_info=None, pad=0
) -> np.ndarray:
    """Wrapper around detect_events that returns 2D np.ndarray for use in predict_oom."""
    event_indices, event_confidence = event_utils.detect_events(event_probability, thres, min_dist, index)

    if block_info is not None:
        # add offset from position of chunk in
        chunk_start_index = block_info[None]["array-location"][0][0]
        chunk_end_index = block_info[None]["array-location"][0][1]

        event_indices += chunk_start_index
        # correct for overlap in all but the first chunk
        if chunk_start_index > 0:
            event_indices -= pad

        # delete detections in overlapping samples
        good_events = np.logical_and(event_indices >= chunk_start_index, event_indices < chunk_end_index)
        if len(good_events):
            event_confidence = event_confidence[good_events]
            event_indices = event_indices[good_events]
        else:
            event_confidence = []
            event_indices = []
    return np.stack((event_indices, event_confidence), axis=1)


def predict_events(
    class_probabilities: np.ndarray,
    samplerate: float = 1.0,
    event_dims: Optional[Iterable[int]] = None,
    event_names: Optional[Iterable[str]] = None,
    event_thres: float = 0.5,
    events_offset: float = 0,
    event_dist: float = 100,
    event_dist_min: float = 0,
    event_dist_max: float = np.inf,
) -> Dict[str, Any]:
    """[summary]

    Args:
        class_probabilities (np.ndarray): [samples, classes][description]
        samplerate (float, optional): Hz
        event_dims (List[int], optional): [description]. Defaults to np.arange(1, nb_classes).
        event_names ([type], optional): [description]. Defaults to event_dims.
        event_thres (float, optional): [description]. Defaults to 0.5.
        events_offset (float, optional): . Defaults to 0 seconds.
        event_dist (float, optional): minimal distance between events for detection (in seconds). Defaults to 100 seconds.
        event_dist_min (float, optional): minimal distance to nearest event for post detection interval filter (in seconds).
                                          Defaults to 0 seconds.
        event_dist_max (float, optional): maximal distance to nearest event for post detection interval filter (in seconds).
                                          Defaults to None (no upper limit).

    Raises:
        ValueError: [description]

    Returns:
        Dict[str, Any]
    """
    if event_dims is None:
        nb_classes = class_probabilities.shape[1]
        event_dims = np.arange(1, nb_classes)

    if event_names is None:
        event_names = [str(d) for d in event_dims]

    if event_dist_max is None:
        event_dist_max = np.inf

    events: Dict[str, Any] = dict()
    if len(event_dims):
        events["samplerate_Hz"] = samplerate
        events["index"] = event_dims
        events["names"] = event_names

        events["seconds"] = []
        events["probabilities"] = []
        events["sequence"] = []

        pad = int(event_dist * samplerate + 1)
        for event_dim, event_name in zip(event_dims, event_names):

            event_indices_and_probs = da.map_overlap(
                _detect_events_oom,
                class_probabilities,
                thres=event_thres,
                min_dist=event_dist * samplerate,
                index=event_dim,
                pad=pad,
                depth=(pad, 0),
                boundary="none",
                trim=False,
                dtype=int,
                meta=np.array(()),
            )
            with ProgressBar(minimum=5):
                event_indices_and_probs = da.compute(event_indices_and_probs)
            event_indices, event_probabilities = event_indices_and_probs[0][:, 0], event_indices_and_probs[0][:, 1]

            events_seconds = event_indices.astype(float) / samplerate
            events_seconds += events_offset

            good_event_indices = event_utils.event_interval_filter(events_seconds, event_dist_min, event_dist_max)
            events["seconds"].extend(events_seconds[good_event_indices])
            events["probabilities"].extend(event_probabilities[good_event_indices])
            events["sequence"].extend([event_name for _ in events_seconds[good_event_indices]])

    return events


def predict_song(
    class_probabilities: np.ndarray,
    params: Dict[str, Any],
    event_thres: float = 0.5,
    event_dist: float = 0.01,
    event_dist_min: float = 0,
    event_dist_max: float = np.inf,
    segment_ref_onsets: Optional[List[float]] = None,
    segment_ref_offsets: Optional[List[float]] = None,
    segment_thres: float = 0.5,
    segment_minlen: float = None,
    segment_fillgap: float = None,
):

    samplerate = params["samplerate_x_Hz"]
    events_offset = 0

    segment_dims = np.where([val == "segment" for val in params["class_types"]])[0]
    segment_names = [str(params["class_names"][segment_dim]) for segment_dim in segment_dims]
    segments = predict_segments(
        class_probabilities,
        samplerate,
        segment_dims,
        segment_names,
        segment_ref_onsets,
        segment_ref_offsets,
        segment_thres,
        segment_minlen,
        segment_fillgap,
    )

    event_dims = np.where([val == "event" for val in params["class_types"]])[0]
    event_names = [str(params["class_names"][event_dim]) for event_dim in event_dims]
    events = predict_events(
        class_probabilities,
        samplerate,
        event_dims,
        event_names,
        event_thres,
        events_offset,
        event_dist,
        event_dist_min,
        event_dist_max,
    )
    return events, segments


def predict(
    x: np.ndarray,
    model_save_name: str = None,
    verbose: int = 1,
    batch_size: int = None,
    model: models.keras.models.Model = None,
    params: Dict = None,
    event_thres: float = 0.5,
    event_dist: float = 0.01,
    event_dist_min: float = 0,
    event_dist_max: float = np.inf,
    segment_thres: float = 0.5,
    segment_use_optimized: bool = True,
    segment_minlen: float = None,
    segment_fillgap: float = None,
    pad: bool = True,
    prepend_data_padding: bool = True,
    save_memory: bool = False,
):
    """[summary]

    Usage:
    Calling predict with the path to the model will load the model and the
    associated params and run inference:
    `das.predict.predict(x=data, model_save_name='tata')`

    To re-use the same model with multiple recordings, load the modal and params
    once and pass them to `predict`
    ```my_model, my_params = das.utils.load_model_and_params(model_save_name)
    for data in data_list:
        das.predict.predict(x=data, model=my_model, params=my_params)
    ```

    Args:
        x (np.array): Audio data [samples, channels]
        model_save_name (str): path with the trunk name of the model. Defaults to None.
        model (keras.model.Models): Defaults to None.
        params (dict): Defaults to None.

        verbose (int): display progress bar during prediction. Defaults to 1.
        batch_size (int): number of chunks processed at once . Defaults to None (the default used during training).
                         Larger batches lead to faster inference.
                         Limited by memory size, in particular for GPUs which typically have 8GB.
                         Large batch sizes lead to loss of samples since only complete batches are used.
        pad (bool): Append zeros to fill up batch. Otherwise the end can be cut.
                    Defaults to False

        event_thres (float): Confidence threshold for detecting peaks. Range 0..1. Defaults to 0.5.
        event_dist (float): Minimal distance between adjacent events during thresholding.
                            Prevents detecting duplicate events when the confidence trace is a little noisy.
                            Defaults to 0.01.
        event_dist_min (float): MINimal inter-event interval for the event filter run during post processing.
                                Defaults to 0.
        event_dist_max (float): MAXimal inter-event interval for the event filter run during post processing.
                                Defaults to np.inf.

        segment_thres (float): Confidence threshold for detecting segments. Range 0..1. Defaults to 0.5.
        segment_use_optimized (bool): Use minlen and fillgap values from param file if they exist.
                                      If segment_minlen and segment_fillgap are provided,
                                      then they will override the values from the param file.
                                      Defaults to True.
        segment_minlen (float): Minimal duration in seconds of a segment used for filtering out spurious detections.
                                Defaults to None.
        segment_fillgap (float): Gap in seconds between adjacent segments to be filled. Useful for correcting brief lapses.
                                 Defaults to None.

        pad (bool): prepend values (repeat last sample value) to fill the last batch.
                    Otherwise, the end of the data will not be annotated because the last, non-full batch will be skipped.
        prepend_data_padding (bool, optional): Restores samples that are ignored
                    in the beginning of the first and the end of the last chunk
                    because of "ignore_boundaries". Defaults to True.
        save_memory (bool): If true, will return memmaped dask.arrays that reside on disk for chunked computations.
                            Convert to np.arrays via the array's `compute()` function.
                            Defaults to False.
    Raises:
        ValueError: [description]

    Returns:
        events: [description]
        segments: [description]
        class_probabilities (np.array): [T, nb_classes]
        class_names (List[str]): [nb_classes]
    """

    if model_save_name is not None:
        model, params = utils.load_model_and_params(model_save_name)
    else:
        assert isinstance(model, models.keras.models.Model)
        assert isinstance(params, dict)

    # use postprocessing values from params and/or args
    if segment_use_optimized and "post_opt" in params and isinstance(params["post_opt"], dict):
        if segment_minlen is None:
            segment_minlen = params["post_opt"]["min_len"]
        if segment_fillgap is None:
            segment_fillgap = params["post_opt"]["gap_dur"]

    if batch_size is not None:
        params["batch_size"] = batch_size

    if pad:
        # figure out length in multiples of batches

        batch_len = params["batch_size"] * params["nb_hist"] + params["nb_hist"]
        x_len_original = len(x)
        pad_len = 0
        if np.remainder(len(x), batch_len) > 0:
            pad_len = batch_len - np.remainder(len(x), batch_len)
        # pad with end val to fill
        x = np.pad(x, ((0, pad_len), (0, 0)), mode="edge")

    class_probabilities = predict_probabilities(x, model, params, verbose, prepend_data_padding)

    if pad:
        # set all song probs in padded section to zero to avoid out of bounds detections!
        # assumes that the non-song class is at index 0
        class_probabilities[-pad_len:, 1:] = 0
        class_probabilities[-pad_len:, 0] = 1
        # trim probs to original len of x
        class_probabilities = class_probabilities[:x_len_original, :]

    events, segments = predict_song(
        class_probabilities=class_probabilities,
        params=params,
        event_thres=event_thres,
        event_dist=event_dist,
        event_dist_min=event_dist_min,
        event_dist_max=event_dist_max,
        segment_ref_onsets=None,
        segment_ref_offsets=None,
        segment_thres=segment_thres,
        segment_minlen=segment_minlen,
        segment_fillgap=segment_fillgap,
    )
    if not save_memory:
        segments["probabilities"] = _to_np(segments["probabilities"])
        segments["samples"] = _to_np(segments["samples"])
        class_probabilities = _to_np(class_probabilities)
    return events, segments, class_probabilities, params["class_names"]


def _to_np(array):
    if isinstance(array, dask.array.core.Array):
        array = array.compute()
    return array


def cli_predict(
    path: str,
    model_save_name: str,
    *,
    save_filename: Optional[str] = None,
    save_format: str = "csv",
    verbose: int = 1,
    batch_size: Optional[int] = None,
    event_thres: float = 0.5,
    event_dist: float = 0.01,
    event_dist_min: float = 0,
    event_dist_max: float = np.inf,
    segment_thres: float = 0.5,
    segment_use_optimized: bool = True,
    segment_minlen: Optional[float] = None,
    segment_fillgap: Optional[float] = None,
    bandpass_low_freq: float = None,
    bandpass_up_freq: float = None,
    resample: bool = True,
):
    """Predict song labels for a wav file or a folder of wav files.

    Saves hdf5 files with keys: events, segments, class_probabilities
    OR csv files with columns: label/start_seconds/stop_seconds

    Args:
        path (str): Path to a single WAV file with the audio data or to a folder with WAV files.
        model_save_name (str): Stem of the path for the model (and parameters).
                               File to load will be MODEL_SAVE_NAME + _model.h5.
        save_filename (Optional[str]): Path to save annotations to.
                                       If omitted, will construct save_filename by stripping the extension
                                       from recording_filename and adding '_das.h5' or '_annotations.csv'.
                                       Will be ignored if `path` is a folder.
        save_format (str): 'csv' or 'h5'.
                           csv: tabular text file with label, start and end seconds for each predicted song.
                           h5: same information as in csv plus confidence values for each sample and song type.
                           Defaults to 'csv'.
        verbose (int): Display progress bar during prediction. Defaults to 1.
        batch_size (Optional[int]): Number of chunks processed at once.
                                    Defaults to None (the default used during training).

        event_thres (float): Confidence threshold for detecting events. Range 0..1. Defaults to 0.5.
        event_dist (float): Minimal distance between adjacent events during thresholding.
                            Prevents detecting duplicate events when the confidence trace is a little noisy.
                            Defaults to 0.01.
        event_dist_min (float): MINimal inter-event interval for the event filter run during post processing.
                                Defaults to 0.
        event_dist_max (float): MAXimal inter-event interval for the event filter run during post processing.
                                          Defaults to np.inf.

        segment_thres (float): Confidence threshold for detecting segments. Range 0..1. Defaults to 0.5.
        segment_use_optimized (bool): Use minlen and fillgap values from param file if they exist.
                                      If segment_minlen and segment_fillgap are provided,
                                      then they will override the values from the param file.
                                      Defaults to True.
        segment_minlen (Optional[float]): Minimal duration of a segment used for filtering out spurious detections.
                                          Defaults to None (keep all segments).
        segment_fillgap (Optional[float]): Gap between adjacent segments to be filled. Useful for correcting brief lapses.
                                           Defaults to None (do not fill gaps).

        bandpass_low_freq (float): Lower cutoff frequency in Hz for bandpass filtering audio data. Defaults to 1.0.
        bandpass_up_freq (float): Upper cutoff frequency in Hz for bandpass filtering audio data. Defaults to samplingrate / 2.

        resample (bool): Resample audio data to the rate expected by the model. Defaults to True.

    Raises:
        ValueError on unknown save_format
    """
    if not (save_format == "csv" or save_format == "h5"):
        raise ValueError(f"Unknown save_format '{save_format}'. Should be either 'csv' or 'h5'.")

    if os.path.isdir(path) and save_filename is not None:
        logging.warning(f"{path} is a folder. Will ignore save_filename argument {save_filename}.")

    # if path is folder: glob contents - all files
    if os.path.isdir(path):
        filenames = glob.glob(f"{path}/*.wav")
        filenames = [filename for filename in filenames if not os.path.isdir(filename)]
    elif os.path.isfile(path):
        filenames = [path]

    logging.info(f"Loading model from {model_save_name}.")
    model, params = utils.load_model_and_params(model_save_name)
    fs_model = params["samplerate_x_Hz"]

    for recording_filename in filenames:
        logging.info(f"   Loading data from {recording_filename}.")
        try:
            # else if path is file - predict only on file but make it single-item list
            x, fs_audio = librosa.load(recording_filename, sr=None, mono=False)
            x = x.T  # [channels, time] -> [time, channels]
            if x.ndim == 1:
                x = x[:, np.newaxis]

            if bandpass_low_freq is not None or bandpass_up_freq is not None:
                x = utils.bandpass_filter_song(x, fs_audio, bandpass_low_freq, bandpass_up_freq)

            if resample and fs_audio != fs_model:
                logging.info(f"   Resampling. Audio rate is {fs_audio}Hz but model was trained on data with {fs_model}Hz.")
                x = utils.resample(x, fs_audio, fs_model)

            logging.info(f"   Annotating using model at {model_save_name}.")
            # TODO: load model once, provide as direct arg
            events, segments, class_probabilities, class_names = predict(
                x,
                None,
                verbose,
                batch_size,
                model,
                params,
                event_thres,
                event_dist,
                event_dist_min,
                event_dist_max,
                segment_thres,
                segment_use_optimized,
                segment_minlen,
                segment_fillgap,
                save_memory=True,
            )

            if "event" in params["class_types"]:
                logging.info(f"   found {len(events['seconds'])} instances of events '{list(set(events['sequence']))}'.")
            if "segment" in params["class_types"]:
                logging.info(
                    f"   found {len(segments['onsets_seconds'])} instances of segments '{list(set(segments['sequence']))}'."
                )

            if save_format == "h5":
                # turn events and segments into df!
                d = {
                    "events": events,
                    "segments": segments,
                    "class_probabilities": class_probabilities,
                    "class_names": class_names,
                }
                if save_filename is None:
                    save_filename = os.path.splitext(recording_filename)[0] + "_das.h5"
                logging.info(f"   Saving results to {save_filename}.")
                flammkuchen.save(save_filename, d)
                logging.info("Done.")
            elif save_format == "csv":
                evt = annot.Events.from_predict(events, segments)
                if save_filename is None:
                    save_filename = os.path.splitext(recording_filename)[0] + "_annotations.csv"
                logging.info(f"   Saving results to {save_filename}.")
                evt.to_df().to_csv(save_filename)
                logging.info("Done.")

            # reset
            if os.path.isdir(path):
                save_filename = None
        except Exception:
            logging.exception(f"Error processing file {recording_filename}.")



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from tensorflow.keras.layers import Layer
from tensorflow.keras import backend as K
from . import backend


# Todo: Filterbank(); init with mel, log, linear, etc.
# not parameterised, just a matrix multiplication


class Filterbank(Layer):
    """
    ### `Filterbank`

    `kapre.filterbank.Filterbank(n_fbs, trainable_fb, sr=None, init='mel', fmin=0., fmax=None,
                                 bins_per_octave=12, image_data_format='default', **kwargs)`

    #### Notes
        Input/output are 2D image format.
        E.g., if channel_first,
            - input_shape: ``(None, n_ch, n_freqs, n_time)``
            - output_shape: ``(None, n_ch, n_mels, n_time)``


    #### Parameters
    * n_fbs: int
       - Number of filterbanks

    * sr: int
        - sampling rate. It is used to initialize ``freq_to_mel``.

    * init: str
        - if ``'mel'``, init with mel center frequencies and stds.

    * fmin: float
        - min frequency of filterbanks.
        - If `init == 'log'`, fmin should be > 0. Use `None` if you got no idea.

    * fmax: float
        - max frequency of filterbanks.
        - If `init == 'log'`, fmax is ignored.

    * trainable_fb: bool,
        - Whether the filterbanks are trainable or not.

    """

    def __init__(
        self,
        n_fbs,
        trainable_fb,
        sr=None,
        init="mel",
        fmin=0.0,
        fmax=None,
        bins_per_octave=12,
        image_data_format="default",
        **kwargs
    ):
        """TODO: is sr necessary? is fmax necessary? init with None?"""
        self.supports_masking = True
        self.n_fbs = n_fbs
        assert init in ("mel", "log", "linear", "uni_random")
        if fmax is None:
            self.fmax = sr / 2.0
        else:
            self.fmax = fmax
        if init in ("mel", "log"):
            assert sr is not None

        self.fmin = fmin
        self.init = init
        self.bins_per_octave = bins_per_octave
        self.sr = sr
        self.trainable_fb = trainable_fb
        assert image_data_format in ("default", "channels_first", "channels_last")
        if image_data_format == "default":
            self.image_data_format = K.image_data_format()
        else:
            self.image_data_format = image_data_format
        super(Filterbank, self).__init__(**kwargs)

    def build(self, input_shape):
        if self.image_data_format == "channels_first":
            self.n_ch = input_shape[1]
            self.n_freq = input_shape[2]
            self.n_time = input_shape[3]
        else:
            self.n_ch = input_shape[3]
            self.n_freq = input_shape[1]
            self.n_time = input_shape[2]

        if self.init == "mel":
            self.filterbank = K.variable(
                backend.filterbank_mel(
                    sr=self.sr, n_freq=self.n_freq, n_mels=self.n_fbs, fmin=self.fmin, fmax=self.fmax
                ).transpose(),
                dtype=K.floatx(),
            )
        elif self.init == "log":
            self.filterbank = K.variable(
                backend.filterbank_log(
                    sr=self.sr, n_freq=self.n_freq, n_bins=self.n_fbs, bins_per_octave=self.bins_per_octave, fmin=self.fmin
                ).transpose(),
                dtype=K.floatx(),
            )

        if self.trainable_fb:
            self.trainable_weights.append(self.filterbank)
        else:
            self.non_trainable_weights.append(self.filterbank)
        super(Filterbank, self).build(input_shape)
        self.built = True

    def compute_output_shape(self, input_shape):
        if self.image_data_format == "channels_first":
            return input_shape[0], self.n_ch, self.n_fbs, self.n_time
        else:
            return input_shape[0], self.n_fbs, self.n_time, self.n_ch

    def call(self, x):
        # reshape so that the last axis is freq axis
        if self.image_data_format == "channels_first":
            x = K.permute_dimensions(x, [0, 1, 3, 2])
        else:
            x = K.permute_dimensions(x, [0, 3, 2, 1])
        output = K.dot(x, self.filterbank)
        # reshape back
        if self.image_data_format == "channels_first":
            return K.permute_dimensions(output, [0, 1, 3, 2])
        else:
            return K.permute_dimensions(output, [0, 3, 2, 1])

    def get_config(self):
        config = {
            "n_fbs": self.n_fbs,
            "sr": self.sr,
            "init": self.init,
            "fmin": self.fmin,
            "fmax": self.fmax,
            "bins_per_octave": self.bins_per_octave,
            "trainable_fb": self.trainable_fb,
        }
        base_config = super(Filterbank, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



"""Code for training networks."""
import time
import logging
import flammkuchen as fl
import numpy as np
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau, TensorBoard
from tensorflow import keras
import os
import yaml
import dask.array as da
from typing import List, Optional, Tuple, Dict, Any, Union
from . import data, models, utils, predict, io, evaluate, tracking, data_hash, augmentation, postprocessing  # , timeseries

logger = logging.getLogger(__name__)

try:  # fixes cuDNN error when using LSTM layer
    import tensorflow as tf

    physical_devices = tf.config.list_physical_devices("GPU")
    if physical_devices:
        for device in physical_devices:
            tf.config.experimental.set_memory_growth(device, enable=True)
except Exception as e:
    logger.exception(e)


def train(
    *,
    data_dir: str,
    x_suffix: str = "",
    y_suffix: str = "",
    save_dir: str = "./",
    save_prefix: Optional[str] = None,
    save_name: Optional[str] = None,
    model_name: str = "tcn",
    nb_filters: int = 16,
    kernel_size: int = 16,
    nb_conv: int = 3,
    use_separable: List[bool] = False,
    nb_hist: int = 1024,
    ignore_boundaries: bool = True,
    batch_norm: bool = True,
    nb_pre_conv: int = 0,
    pre_nb_conv: int = 3,
    pre_nb_dft: int = 64,
    pre_kernel_size: int = 3,
    pre_nb_filters: int = 16,
    upsample: bool = True,
    dilations: Optional[List[int]] = None,
    nb_lstm_units: int = 0,
    verbose: int = 2,
    batch_size: int = 32,
    nb_epoch: int = 400,
    learning_rate: Optional[float] = None,
    reduce_lr: bool = False,
    reduce_lr_patience: int = 5,
    fraction_data: Optional[float] = None,
    first_sample_train: Optional[int] = 0,
    last_sample_train: Optional[int] = None,
    first_sample_val: Optional[int] = 0,
    last_sample_val: Optional[int] = None,
    seed: Optional[int] = None,
    batch_level_subsampling: bool = False,
    augmentations: Optional[str] = None,
    tensorboard: bool = False,
    wandb_api_token: Optional[str] = None,
    wandb_project: Optional[str] = None,
    wandb_entity: Optional[str] = None,
    log_messages: bool = False,
    nb_stacks: int = 2,
    with_y_hist: bool = True,
    balance: bool = False,
    version_data: bool = True,
    post_opt: bool = False,
    post_opt_nb_workers: int = -1,
    post_opt_fill_gaps_min: float = 0.0005,
    post_opt_fill_gaps_max: float = 1.0,
    post_opt_fill_gaps_steps: int = 20,
    post_opt_min_len_min: float = 0.0005,
    post_opt_min_len_max: float = 1.0,
    post_opt_min_len_steps: int = 20,
    resnet_compute: bool = False,
    resnet_train: bool = False,
    _qt_progress: bool = False,
) -> Tuple[keras.Model, Dict[str, Any], keras.callbacks.History]:
    """Train a DAS network.

    Args:
        data_dir (str): Path to the directory or file with the dataset for training.
                        Accepts npy-dirs (recommended), h5 files or zarr files.
                        See documentation for how the dataset should be organized.
        x_suffix (str): Select dataset used for training in the data_dir by suffix (`y_` + X_SUFFIX).
                        Defaults to '' (will use the standard data 'x')
        y_suffix (str): Select dataset used as a training target in the data_dir by suffix (`y_` + Y_SUFFIX).
                        Song-type specific targets can be created with a training dataset,
                        Defaults to '' (will use the standard target 'y')
        save_dir (str): Directory to save training outputs.
                        The path of output files will constructed from the SAVE_DIR, an optional SAVE_PREFIX,
                        and the time stamp of the start of training.
                        Defaults to the current directory ('./').
        save_prefix (Optional[str]): Prepend to timestamp.
                           Name of files created will be start with SAVE_DIR/SAVE_PREFIX + "_" + TIMESTAMP
                           or with SAVE_DIR/TIMESTAMP if SAVE_PREFIX is empty.
                           Defaults to '' (empty).
        save_name (Optional[str]): Append to prefix.
                           Name of files created will be start with SAVE_DIR/SAVE_PREFIX + "_" + SAVE_NAME
                           or with SAVE_DIR/SAVE_NAME if SAVE_PREFIX is empty.
                           Defaults to the timestamp YYYYMMDD_hhmmss.
        model_name (str): Network architecture to use.
                          See das.models for a description of all models.
                          Defaults to `tcn`.
        nb_filters (int): Number of filters per layer.
                          Defaults to 16.
        kernel_size (int): Duration of the filters (=kernels) in samples.
                           Defaults to 16.
        nb_conv (int): Number of TCN blocks in the network.
                       Defaults to 3.
        use_separable (List[bool]): Specify which TCN blocks should use separable convolutions.
                                    Provide as a space-separated sequence of "False" or "True.
                                    For instance: "True False False" will set the first block in a
                                    three-block (as given by nb_conv) network to use separable convolutions.
                                    Defaults to False (no block uses separable convolutions).
        nb_hist (int): Number of samples processed at once by the network (a.k.a chunk duration).
                       Defaults to 1024 samples.
        ignore_boundaries (bool): Minimize edge effects by discarding predictions at the edges of chunks.
                                  Defaults to True.
        batch_norm (bool): Batch normalize.
                           Defaults to True.
        nb_pre_conv (int): Adds fronted with downsampling. The downsampling factor is `2**nb_pre_conv`.
                           The type of frontend depends on the model:
                           if model is `tcn`: adds a frontend of N conv blocks (conv-relu-batchnorm-maxpool2) to the TCN.
                           if model is `tcn_tcn`: adds a frontend of N TCN blocks to the TCN.
                           if model is `tcn_stft`: adds a trainable STFT frontend.
                           Defaults to 0 (no frontend, no downsampling).
        pre_nb_dft (int): Duration of filters (in samples) for the STFT frontend.
                          Number of filters is pre_nb_dft // 2 + 1.
                          Defaults to 64.
        pre_nb_filters (int): Number of filters per layer in the pre-processing TCN.
                              Defaults to 16.
        pre_kernel_size (int): Duration of filters (=kernels) in samples in the pre-processing TCN.
                               Defaults to 3.
        upsample (bool): whether or not to restore the model output to the input samplerate.
                         Should generally be True during training and evaluation but my speed up inference.
                         Defaults to True.
        dilations (List[int]): List of dilation rate, defaults to [1, 2, 4, 8, 16] (5 layer with 2x dilation per TCN block)
        nb_lstm_units (int): If >0, adds LSTM with `nb_lstm_units` LSTM units to the output of the stack of TCN blocks.
                             Defaults to 0 (no LSTM layer).
        verbose (int): Verbosity of training output (0 - no output during training, 1 - progress bar, 2 - one line per epoch).
                       Defaults to 2.
        batch_size (int): Batch size
                          Defaults to 32.
        nb_epoch (int): Maximal number of training epochs.
                        Training will stop early if validation loss did not decrease in the last 20 epochs.
                        Defaults to 400.
        learning_rate (Optional[float]): Learning rate of the model. Defaults should work in most cases.
                               Values typically range between 0.1 and 0.00001.
                               If None, uses model specific defaults: `tcn` 0.0001, `tcn_stft` and `tcn_tcn` 0.0005.
                               Defaults to None.
        reduce_lr (bool): Reduce learning rate when the validation loss plateaus.
                          Defaults to False.
        reduce_lr_patience (int): Number of epochs w/o a reduction in validation loss after which
                                  to trigger a reduction in learning rate.
                                  Defaults to 5 epochs.
        fraction_data (Optional[float]): Fraction of training and validation data to use.
                                         Defaults to 1.0.
                                         Overriden by setting all four *_sample_* args.
        first_sample_train (Optional[int]): Defaults to 0 (first sample in training dataset).
                                            Note 1: all four *_sample_* args must be set - otherwise they will be ignored.
                                            Note 2: Overrides fraction_data.
        last_sample_train (Optional[int]): Defaults to None (use last sample in training dataset).
        first_sample_val (Optional[int]): Defaults to 0 (first sample in validation dataset).
        last_sample_val (Optional[int]): Defaults to None (use last sample in validation dataset).
        seed (Optional[int]): Random seed to reproducibly select fractions of the data.
                              Defaults to None (no seed).
        batch_level_subsampling (bool): Select fraction of data for training from random subset of shuffled batches.
                                        If False, select a continuous chunk of the recording.
                                        Defaults to False.
        augmentations (Optional[str]): Path to yaml file or dictionary with the specification of augmentations.
                                       Defaults to None (no augmentations).
        tensorboard (bool): Write tensorboard logs to save_dir. Defaults to False.
        wandb_api_token (Optional[str]): API token for logging to wandb.
                                           Defaults to None (no logging to wandb).
        wandb_project (Optional[str]): Project to log to for wandb.
                                         Defaults to None (no logging to wandb).
        wandb_entity (Optional[str]): Entity to log to for wandb.
                                        Defaults to None (no logging to wandb).
        log_messages (bool): Sets terminal logging level to INFO.
                             Defaults to False (will follow existing settings).

        nb_stacks (int): Unused if model name is `tcn`, `tcn_tcn`, or `tcn_stft`. Defaults to 2.
        with_y_hist (bool): Unused if model name is `tcn`, `tcn_tcn`, or `tcn_stft`. Defaults to True.
        balance (bool): Balance data. Weights class-wise errors by the inverse of the class frequencies.
                        Defaults to False.
        version_data (bool): Save MD5 hash of the data_dir to log and params.yaml.
                             Defaults to True (set to False for large datasets since it can be slow).

        post_opt (bool): Optimize post processing (delete short detections, fill brief gaps).
                        Defaults to False.
        post_opt_nb_workers (int): Number of parallel processes during post_opt. Defaults to -1 (same number as cores).
        post_opt_fill_gaps_min (float): Defaults to 0.0005 seconds.
        post_opt_fill_gaps_max (float): Defaults to 1 second.
        post_opt_fill_gaps_steps (int): Defaults to 20.
        post_opt_min_len_min (float): Defaults to 0.0005 seconds.
        post_opt_min_len_max (float): Defaults to 1 second.
        post_opt_min_len_steps (int): Defaults to 20.

        resnet_compute (bool): Defaults to False.
        resnet_train (bool): Defaults to False.
        Returns
            model (keras.Model)
            params (Dict[str, Any])
            history (keras.callbacks.History)
    """
    # _qt_progress: tuple of (multiprocessing.Queue, threading.Event)
    #        The queue is used to transmit progress updates to the GUI,
    #        the event is set in the GUI to stop training.
    if log_messages:
        logging.basicConfig(level=logging.INFO)

    if dilations is None:
        dilations = [1, 2, 4, 8, 16]

    # FIXME THIS IS NOT GREAT:
    sample_weight_mode = None
    data_padding = 0
    if with_y_hist:  # regression
        return_sequences = True
        stride = nb_hist
        y_offset = 0
        sample_weight_mode = "temporal"
        if ignore_boundaries:
            data_padding = int(
                np.ceil(kernel_size * nb_conv)
            )  # this does not completely avoid boundary effects but should minimize them sufficiently
            stride = stride - 2 * data_padding
    else:  # classification
        return_sequences = False
        stride = 1  # should take every sample, since sampling rates of both x and y are now the same
        y_offset = int(round(nb_hist / 2))
        upsample = False

    if stride <= 0:
        raise ValueError(
            "Stride <=0 - needs to be >0. Possible solutions: reduce kernel_size, increase nb_hist parameters, uncheck ignore_boundaries"
        )

    if not upsample:
        output_stride = int(2 ** nb_pre_conv)
    else:
        output_stride = 1  # since we upsample output to original sampling rate. w/o upsampling: `output_stride = int(2**nb_pre_conv)` since each pre-conv layer does 2x max pooling

    if save_prefix is None:
        save_prefix = ""

    if len(save_prefix):
        save_prefix = save_prefix + "_"
    params = locals()
    del params["_qt_progress"]

    # remove learning rate param if not set so the value from the model def is used
    if params["learning_rate"] is None:
        del params["learning_rate"]

    if "_multi" in model_name:
        params["unpack_channels"] = True

    logger.info(f"Loading data from {data_dir}.")
    d = io.load(data_dir, x_suffix=x_suffix, y_suffix=y_suffix)

    params.update(d.attrs)  # add metadata from data.attrs to params for saving

    if version_data:
        logger.info("Versioning the data:")
        params["data_hash"] = data_hash.hash_data(data_dir)
        logger.info(f"   MD5 hash of {data_dir} is")
        logger.info(f"   {params['data_hash']}")

    sample_bounds_provided = (
        first_sample_train is not None
        and last_sample_train is not None
        and first_sample_val is not None
        and last_sample_val is not None
    )

    if fraction_data is not None and not sample_bounds_provided:
        if fraction_data > 1.0:  # seconds
            logger.info(
                f"{fraction_data} seconds corresponds to {fraction_data / (d['train']['x'].shape[0] / d.attrs['samplerate_x_Hz']):1.4f} of the training data."
            )
            fraction_data = np.min((fraction_data / (d["train"]["x"].shape[0] / d.attrs["samplerate_x_Hz"]), 1.0))
        elif fraction_data < 1.0:
            logger.info(f"Using {fraction_data:1.4f} of the training and validation data.")

    if (
        fraction_data is not None and not batch_level_subsampling and not sample_bounds_provided and fraction_data != 1.0
    ):  # train on a subset
        min_nb_samples = nb_hist * (batch_size + 2)  # ensure the generator contains at least one full batch
        first_sample_train, last_sample_train = data.sub_range(
            d["train"]["x"].shape[0], fraction_data, min_nb_samples, seed=seed
        )
        first_sample_val, last_sample_val = data.sub_range(d["val"]["x"].shape[0], fraction_data, min_nb_samples, seed=seed)
    elif sample_bounds_provided:
        logger.info("Using provided start/end samples:")
        logger.info(f"Train: {first_sample_train}:{last_sample_train}, Val: {first_sample_val}:{last_sample_val}.")

    # TODO clarify nb_channels, nb_freq semantics - always [nb_samples,..., nb_channels] -  nb_freq is ill-defined for 2D data
    params.update(
        {
            "nb_freq": d["train"]["x"].shape[1],
            "nb_channels": d["train"]["x"].shape[-1],
            "nb_classes": len(params["class_names"]),
            "first_sample_train": first_sample_train,
            "last_sample_train": last_sample_train,
            "first_sample_val": first_sample_val,
            "last_sample_val": last_sample_val,
        }
    )
    logger.info("Parameters:")
    logger.info(params)

    logger.info("Preparing data")
    if fraction_data is not None and batch_level_subsampling:  # train on a subset
        np.random.seed(seed)
        shuffle_subset = fraction_data
    else:
        shuffle_subset = None

    if augmentations:
        if isinstance(augmentations, str):
            logger.info(f"Initializing augmentations from {augmentations}.")
            aug_params = yaml.safe_load(open(augmentations, "r"))
        else:
            aug_params = augmentations
        params["augmentations"] = aug_params
        augs = augmentation.Augmentations.from_dict(params["augmentations"])
        logger.info(f"   Got {len(augs)} augmentations.")
    else:
        augs = None

    params["class_weights"] = None
    if balance:
        logger.info("Balancing classes:")
        logger.info("   Computing class weights.")
        params["class_weights"] = data.compute_class_weights(d["train"]["y"][first_sample_train:last_sample_train])
        logger.info(f"   {params['class_weights']}")

    data_gen = data.AudioSequence(
        d["train"]["x"],
        d["train"]["y"],
        shuffle=True,
        shuffle_subset=shuffle_subset,
        first_sample=first_sample_train,
        last_sample=last_sample_train,
        nb_repeats=1,
        batch_processor=augs,
        **params,
    )
    val_gen = data.AudioSequence(
        d["val"]["x"],
        d["val"]["y"],
        shuffle=False,
        shuffle_subset=shuffle_subset,
        first_sample=first_sample_val,
        last_sample=last_sample_val,
        **params,
    )
    # data_gen = timeseries.timeseries_dataset_from_array(d['train']['x'], d['train']['y'],
    #                               sequence_length=params['nb_hist'], sequence_stride=stride,
    #                               shuffle=True, batch_size=batch_size,
    #                               start_index=first_sample_train, end_index=last_sample_train)
    # val_gen = timeseries.timeseries_dataset_from_array(d['val']['x'], d['val']['y'],
    #                              sequence_length=params['nb_hist'], sequence_stride=stride,
    #                              shuffle=False, batch_size=batch_size,
    #                              start_index=first_sample_val, end_index=last_sample_val)

    logger.info(f"Training data:")
    logger.info(f"   {data_gen}")
    logger.info(f"Validation data:")
    logger.info(f"   {val_gen}")

    logger.info("Building network")
    try:
        model = models.model_dict[model_name](**params)
    except KeyError as e:
        logger.exception(e)
        raise ValueError(f"Model name was {model_name} but only {list(models.model_dict)} allowed.")

    logger.info(model.summary())
    os.makedirs(os.path.abspath(save_dir), exist_ok=True)
    if save_name is None:
        save_name = time.strftime("%Y%m%d_%H%M%S")
    save_name = "{0}/{1}{2}".format(save_dir, save_prefix, save_name)
    params["save_name"] = save_name
    logger.info(f"Will save to {save_name}.")

    # SET UP CALLBACKS
    checkpoint_save_name = save_name + "_model.h5"  # this will overwrite intermediates from previous epochs
    callbacks = [
        ModelCheckpoint(checkpoint_save_name, save_best_only=True, save_weights_only=False, monitor="val_loss", verbose=1),
        EarlyStopping(monitor="val_loss", patience=20, verbose=1),
    ]

    if reduce_lr:
        callbacks.append(ReduceLROnPlateau(patience=reduce_lr_patience, verbose=1))

    if _qt_progress:
        callbacks.append(utils.QtProgressCallback(nb_epoch, _qt_progress))

    if tensorboard:
        callbacks.append(TensorBoard(log_dir=save_dir))

    if wandb_api_token and wandb_project:  # could also get those from env vars!
        del params["wandb_api_token"]
        wandb = tracking.Wandb(wandb_project, wandb_api_token, wandb_entity, params)
        if wandb:
            callbacks.append(wandb.callback())

    utils.save_params(params, save_name)

    # TRAIN NETWORK
    logger.info("start training")
    fit_hist = model.fit(
        data_gen,
        epochs=nb_epoch,
        steps_per_epoch=min(len(data_gen), 1000),
        verbose=verbose,
        validation_data=val_gen,
        callbacks=callbacks,
    )

    tf.keras.backend.clear_session()

    # OPTIMIZE POSTPROCESSING
    if post_opt:
        logger.info("OPTIMIZING POSTPROCESSING:")

        gap_durs = np.geomspace(float(post_opt_fill_gaps_min), float(post_opt_fill_gaps_max), int(post_opt_fill_gaps_steps))
        min_lens = np.geomspace(float(post_opt_min_len_min), float(post_opt_min_len_max), int(post_opt_min_len_steps))

        best_gap_dur, best_min_len, scores = postprocessing.optimize(
            dataset_path=data_dir,
            model_save_name=save_name,
            gap_durs=gap_durs,
            min_lens=min_lens,
            nb_workers=post_opt_nb_workers,
        )

        logger.info(f"  Score on training data changed from {scores['train_pre']:1.4} to {scores['train']:1.4}.")
        if scores["val_pre"] is not None:
            logger.info(f"  Score on validation data changed from {scores['val_pre']:1.4} to {scores['val']:1.4}.")
        logger.info("  Optimal parameters for postprocessing:")
        logger.info(f"     gap_dur={best_gap_dur} seconds")
        logger.info(f"     min_len={best_min_len} seconds")

        params["post_opt"] = {
            "gap_dur": best_gap_dur,
            "min_len": best_min_len,
            "score_train": scores["train"],
            "score_val": scores["val"],
        }

        logger.info(f'   Updating params file "{save_name}_params.yaml" with the results.')
        utils.save_params(params, save_name)
        logger.info("DONE")

    # TEST
    # TODO use postprocessing params
    logger.info("TESTING:")
    if "test" not in d or len(d["test"]["x"]) < nb_hist:
        logger.info("   No test data - skipping final evaluation step.")
    else:
        logger.info(f"   Re-loading last best model from {checkpoint_save_name}.")
        model.load_weights(checkpoint_save_name)

        logger.info("   Predicting.")
        x_test, y_test, y_pred = evaluate.evaluate_probabilities(x=d["test"]["x"], y=d["test"]["y"], model=model, params=params)

        labels_test = predict.labels_from_probabilities(y_test)
        labels_pred = predict.labels_from_probabilities(y_pred)

        logger.info("   Evaluating.")
        conf_mat, report = evaluate.evaluate_segments(labels_test, labels_pred, params["class_names"], report_as_dict=True)
        logger.info(conf_mat)
        logger.info(report)
        params["conf_mat"] = conf_mat.tolist()
        params["report"] = report
        logger.info(f'   Updating params file "{save_name}_params.yaml" with the test results.')
        utils.save_params(params, save_name)

        if wandb_api_token and wandb_project:  # could also get those from env vars!
            wandb.log_test_results(report)

        save_filename = "{0}_results.h5".format(save_name)
        logger.info(f"   Saving to {save_filename}.")
        del params["data_splits"]  # paths with '/' break flammkuchen/pytables
        results_dict = {
            "fit_hist": dict(fit_hist.history),
            "confusion_matrix": conf_mat,
            "classification_report": report,
            "x_test": x_test,
            "y_test": y_test,
            # 'y_pred': np.array(y_pred),
            "labels_test": labels_test,
            # 'labels_pred': np.array(labels_pred),
            "params": params,
        }
        fl.save(save_filename, results_dict)
        da.to_hdf5(save_filename, {"/y_pred": y_pred, "/labels_pred": labels_pred})

    logger.info("DONE.")
    return model, params, fit_hist



"""Utilities for handling events."""
import numpy as np
import peakutils
from typing import Iterable, Optional, List, Tuple, Union


def find_nearest(array, values):
    """Find nearest occurrence of each item of values in array.

    Args:
        array: find nearest in this list
        values: queries

    Returns:
        val: nearest val in array to each item in values
        idx: index of nearest val in array to each item in values
        dist: distance to nearest val in array for each item in values
        NOTE: Returns nan-arrays of the same size as values if `array` is empty.
    """
    if len(values) and len(array):  # only do this if both inputs are non-empty lists
        values = np.atleast_1d(values)
        abs_dist = np.abs(np.int64(np.subtract.outer(array, values)))
        idx = abs_dist.argmin(0)
        dist = abs_dist.min(0)
        val = array[idx]
    else:
        idx = np.full_like(values, fill_value=np.nan)
        dist = np.full_like(values, fill_value=np.nan)
        val = np.full_like(values, fill_value=np.nan)
    return val, idx, dist


def detect_events(
    event_probability: np.ndarray, thres: float = 0.5, min_dist: int = 100, index: int = 0
) -> Tuple[np.ndarray, np.ndarray]:
    """Detect events as peaks in probabilitiy.

    Args:
        event_probability ([np.ndarray]): [T, nb_classes]
        thres (float, optional): [description]. Defaults to 0.5.
        min_dist (int, optional): [description]. Defaults to 100 samples.
        index: (int, Optional): List of indices into axis 1 for which to compute the labels.
                                     Defaults to None (use all indices).
    Returns:
        event_indices: index of each detected event
        event_confidence: event_probability at the event_index
    """
    event_indices = peakutils.indexes(event_probability[:, index], thres=thres, min_dist=min_dist, thres_abs=True)

    if len(event_indices):  # guard against empty event_indices
        event_confidence = event_probability[event_indices, index]
    else:
        event_confidence = []

    return event_indices, event_confidence


def match_events(eventindices_true, eventindices_pred, tol=100):
    """Find events eventindices_pred that match those (within tol) in eventindices_true.

    Args:
        eventindices_true: list of reference event indices
        eventindices_pred: list of detected event indices
        tol: n samples within which events are deemed identical
    Returns:
        nearest_event: masked array copy of eventindices_pred, mask=True indicates entries in pred not closest within tol in true
        nearest_dist: dist of each eventindices_pred to the nearest true_event
    """

    nearest_dist = np.zeros_like(eventindices_pred)
    nearest_event = np.zeros_like(eventindices_pred)
    # find nearest true event for each predicted event
    _, nearest_event, nearest_dist = find_nearest(eventindices_true, eventindices_pred)
    nearest_event = nearest_event.astype(np.float)

    # flag those that have no nearby event
    nearest_event = np.ma.masked_array(nearest_event, mask=nearest_dist > tol)
    if len(eventindices_true) == 0:
        nearest_event.mask = True  # all detections are false positives
    else:
        # flag doublettes - keep only nearest
        for idx in np.unique(nearest_event[nearest_event >= 0]):
            hits = np.where(nearest_event == idx)[0]
            if len(hits) > 1:
                nearest = np.argmin(nearest_dist[nearest_event == idx])  # find closest hit
                hits = np.delete(hits, nearest)
                nearest_event.mask[hits] = True

    return nearest_event.astype(np.uintp), nearest_dist


def event_interval_filter(events: Iterable, event_dist_min: float = 0, event_dist_max: float = np.inf) -> Iterable:
    """[summary]

    Args:
        events (Iterable): Iterable (list, np.array) of event times in seconds.
        event_dist_min (float, optional): [description]. Defaults to 0 second.
        event_dist_max (float, optional): [description]. Defaults to np.inf seconds.

    Returns:
        good_event_indices (Iterable): indices of events to keep `events[good_event_indices]`.
    """

    ipi_pre = np.diff(events, prepend=np.inf)
    ipi_post = np.diff(events, append=np.inf)

    ipi_too_long = np.logical_and(ipi_pre > event_dist_max, ipi_post > event_dist_max)
    ipi_too_short = np.logical_or(ipi_pre < event_dist_min, ipi_post < event_dist_min)
    if len(ipi_too_short):
        ipi_too_short[0] = False  # otherwise first event will always be removed

    good_event_indices = ~np.logical_or(ipi_too_long, ipi_too_short)

    return good_event_indices


def evaluate_eventtimes(eventtimes_true, eventtimes_pred, samplerate, tol=0.01):
    """[summary]

    Args:
        eventtimes_true ([type]): in seconds
        eventtimes_pred ([type]): in seconds
        samplerate (float): in Hz
        tol (int, optional): in seconds [description]. Defaults to 0.01 seconds.

    Returns:
        [type]: [description]
    """

    # match_events works with indices - so need to convert eventtimes to indices and distance back to times
    eventindices_true = eventtimes_true * samplerate
    eventindices_pred = eventtimes_pred * samplerate

    nearest_pred_event, nearest_dist_indices = match_events(eventindices_true, eventindices_pred, tol * samplerate)
    nearest_true_event, _ = match_events(eventindices_pred, eventindices_true, tol * samplerate)
    nearest_dist = nearest_dist_indices / samplerate  # convert back to seconds

    d = dict()
    # pred events that have no nearby true event (or there is another pred event nearer to the true event)
    d["FP"] = np.sum(nearest_pred_event.mask)
    d["TP"] = len(nearest_pred_event[np.isfinite(nearest_dist)].compressed())
    d["FN"] = max(0, np.sum(nearest_true_event.mask))

    if d["FP"] == 0:  # if there are no false positives (even for no detections), then precision is 1.0
        d["precision"] = 1.0
    elif (d["TP"] + d["FP"]) == 0:
        d["precision"] = 0.0
    else:
        d["precision"] = d["TP"] / (d["TP"] + d["FP"])

    if d["FN"] == 0:  # if there are no false negatives (even for no detections), the recall is 1.0 (?!)
        d["recall"] = 1.0
    elif d["TP"] + d["FN"] == 0:
        d["recall"] = 0
    else:
        d["recall"] = d["TP"] / (d["TP"] + d["FN"])

    if (d["precision"] + d["recall"]) == 0:
        d["f1_score"] = 0
    else:
        d["f1_score"] = 2 * (d["precision"] * d["recall"]) / (d["precision"] + d["recall"])

    return d, nearest_pred_event, nearest_true_event, nearest_dist



"""General utilities"""
import tensorflow.keras as keras
import logging
import time
import numpy as np
import yaml
import h5py
import scipy.signal
from . import kapre
from . import tcn
from . import models
from typing import Dict, Callable, Any, List, Tuple, Optional


def load_model(
    file_trunk: str,
    model_dict: Dict[str, Callable],
    model_ext: str = "_model.h5",
    params_ext: str = "_params.yaml",
    compile: bool = True,
    custom_objects: Optional[Dict[str, Callable]] = None,
):
    """Load model with weights.

    First tries to load the full model directly using keras.models.load_model - this will likely fail for models with custom layers.
    Second, try to init model from parameters and then add weights...

    Args:
        file_trunk (str): [description]
        model_dict (Dict[str, Callable): [description]
        model_ext (str, optional): [description]. Defaults to '_weights.h5'.
        params_ext (str, optional): [description]. Defaults to '_params.yaml'.
        compile (bool, optional): [description]. Defaults to True.
        custom_objects (dict, optional): ...

    Returns:
        keras.Model
    """

    if custom_objects is None:
        custom_objects = {"Spectrogram": kapre.time_frequency.Spectrogram, "TCN": tcn.tcn_new.TCN}

    try:
        model_filename = _download_if_url(file_trunk + model_ext)
        model = keras.models.load_model(model_filename, custom_objects=custom_objects)
    except (SystemError, ValueError, AttributeError):
        logging.debug(
            "Failed to load model using keras, likely because it contains custom layers. Will try to init model architecture from code and load weights from `_model.h5` into it.",
            exc_info=False,
        )
        logging.debug("", exc_info=True)
        model = load_model_from_params(file_trunk, model_dict, weights_ext=model_ext, params_ext=params_ext, compile=compile)
    return model


def load_model_from_params(
    file_trunk: str,
    model_dict: Dict[str, Callable],
    weights_ext: str = "_model.h5",
    params_ext: str = "_params.yaml",
    compile: bool = True,
):
    """Init architecture from code and load model weights into it. Helps with model loading issues across TF versions.

    Args:
        file_trunk (str): [description]
        models_dict ([type]): [description]
        weights_ext (str, optional): [description]. Defaults to '_model.h5' (use weights from model file).
        params_ext (str, optional): [description]. Defaults to '_params.yaml'.
        compile (bool, optional): [description]. Defaults to True.

    Returns:
        keras.Model
    """
    params = load_params(file_trunk, params_ext=params_ext)

    # get the model - calls the function that generates a model with parameters
    model = model_dict[params["model_name"]](**params)
    weights_filename = _download_if_url(file_trunk + weights_ext)
    model.load_weights(weights_filename)

    if compile:
        # Compile with random standard optimizer and loss so we can use the model for prediction
        # Just re-compile the model if you want a particular optimizer and loss.
        model.compile(optimizer=keras.optimizers.Adam(amsgrad=True), loss="mean_squared_error")
    return model


def save_params(params: Dict[str, Any], file_trunk: str, params_ext: str = "_params.yaml"):
    """Save model/training parameters to yaml.

    Args:
        params (Dict[str]): [description]
        file_trunk (str): [description]
        params_ext (str, optional): [description]. Defaults to '_params.yaml'.
    """
    with open(file_trunk + params_ext, "w") as f:
        yaml.dump(params, f)


def load_params(file_trunk: str, params_ext: str = "_params.yaml") -> Dict[str, Any]:
    """Load model/training parameters from yaml

    Args:
        file_trunk (str): [description]
        params_ext (strs, optional): [description]. Defaults to '_params.yaml'.

    Returns:
        Dict[str, Any]: Parameter dictionary
    """
    filename = _download_if_url(file_trunk + params_ext)
    with open(filename, "r") as f:
        try:
            params = yaml.unsafe_load(f)
        except AttributeError:
            params = yaml.load(f)
    return params


def load_model_and_params(
    model_save_name, model_dict=models.model_dict, custom_objects=None
) -> Tuple[keras.Model, Dict[str, Any]]:
    """[summary]

    Args:
        model_save_name ([type]): [description]
        model_dict ([type], optional): [description]. Defaults to models.model_dict.
        custom_objects

    Returns:
        keras.Model, Dict[str, Any]: [description]
    """
    params = load_params(model_save_name)
    model = load_model(model_save_name, model_dict=model_dict, custom_objects=custom_objects)
    return model, params


def _download_if_url(url: str):
    if not url.startswith("http"):
        return url
    else:
        import urllib.request
        import tempfile
        from pathlib import Path

        filename = url.split("/")[-1]  # get filename
        tmpdir = tempfile.mkdtemp()
        local_path = Path(tmpdir) / filename
        urllib.request.urlretrieve(url, local_path)
        return local_path


def load_from(filename: str, datasets: List[str]):
    """Load datasets from h5 file.

    Args:
        filename (str)
        datasets (List[str]): Names of the datasets (=keys) to load

    Returns:
        [type]: [description]
    """
    data = dict()
    with h5py.File(filename, "r") as f:
        data = {dataset: f[dataset][:] for dataset in datasets}
    return data


class Timer:
    def __init__(self, verbose=False):
        self.verbose = verbose
        self.start = None
        self.end = None
        self.elapsed = None

    def __enter__(self):
        self.start = time.perf_counter()
        return self

    def __exit__(self, *args):
        self.end = time.perf_counter()
        self.elapsed = self.end - self.start
        if self.verbose:
            print(self)

    def __str__(self):
        if self.start is None:
            s = "Timer not started yet."
        elif self.end is None:
            s = "Timer still running."
        elif self.elapsed is not None:
            s = f"Time elapsed {self.elapsed:1.2f} seconds."
        else:
            s = "Timer in unexpected state."
        return s


class QtProgressCallback(keras.callbacks.Callback):
    def __init__(self, nb_epochs, comms):
        """Init the callback.

        Args:
            nb_epochs ([type]): number of training epochs
            comms (tuple): tuple of (multiprocessing.Queue, threading.Event)
                The queue is used to transmit progress updates to the GUI,
                the event is set in the GUI to stop training.
        """
        super().__init__()
        self.nb_epochs = nb_epochs
        self.queue = comms[0]
        self.stop_event = comms[1]

    def _check_if_stopped(self):
        try:
            if self.stop_event.is_set():
                self.model.stop_training = True
        except Exception as e:
            print(e)

    def on_train_begin(self, logs=None):
        self.queue.put((0, "Starting training."))

    def on_train_end(self, logs=None):
        self.queue.put((-1, "Finishing training."))

    def on_epoch_end(self, epoch, logs=None):
        self.queue.put((epoch, f"Epoch {epoch}/{self.nb_epochs}"))

    def on_train_batch_end(self, batch, logs=None):
        self._check_if_stopped()

    def on_test_batch_end(self, batch, logs=None):
        self._check_if_stopped()

    def on_predict_batch_end(self, batch, logs=None):
        self._check_if_stopped()


def resample(x: np.array, fs_audio: float, fs_model: float):
    """Resample audio to model rate.

    Rounds rates to next even number for efficiency.

    Args:
        x (np.array): _description_
        fs_audio (float): _description_
        fs_model (float): _description_

    Returns:
        np.array: Audio resample to fs_model.
    """
    fs_audio_even = int(fs_audio // 2) * 2
    fs_model_even = int(fs_model // 2) * 2
    gcd = np.gcd(fs_audio_even, fs_model_even)
    x = scipy.signal.resample_poly(x, fs_audio_even // gcd, fs_model_even // gcd, axis=0)
    return x


def bandpass_filter_song(
    x: np.ndarray, sampling_rate_hz: float, f_low: Optional[float] = None, f_high: Optional[float] = None
) -> np.ndarray:
    """Band-pass filter channel data

    Args:
        x (np.ndarray): Audio data[T,] or [T, nb_channels]
        sampling_rate_hz (float): Sampling rate in Hz
        f_low (Optional[float], optional): Lower cutoff in Hz. Defaults to 1.0 (None).
        f_high (Optional[float], optional): Upper cutoff in Hz. Defaults to sampling_rate_hz/2 (None).

    Returns:
        np.ndarray: Sampled data
    """
    if f_low is None:
        f_low = 1.0

    if f_high is None:
        f_high = sampling_rate_hz / 2 - 1

    f_high = min(f_high, sampling_rate_hz / 2 - 1)

    sos_bp = scipy.signal.butter(5, [f_low, f_high], "bandpass", output="sos", fs=sampling_rate_hz)
    x = scipy.signal.sosfiltfilt(sos_bp, x, axis=0)
    return x



import numpy as np
import zarr
from typing import List, Dict, Mapping, Optional
import pandas as pd
import scipy.signal
import logging
import sklearn.model_selection


logger = logging.getLogger(__name__)


def init_store(
    nb_channels: int,
    nb_classes: int,
    samplerate: Optional[float] = None,
    make_single_class_datasets: bool = False,
    class_names: List[str] = None,
    class_types: List[str] = None,
    store_type=zarr.TempStore,
    store_name: str = "store.zarr",
    chunk_len: int = 1_000_000,
):
    """[summary]

    Args:
        nb_channels (int): [description]
        nb_classes (int): [description]  <- should infer from class_names!
        samplerate (float, optional): [description]. Defaults to None.
        make_single_class_datasets (bool, optional): make y_suffix and attrs['class_names/types_suffix']. Defaults to None.
        class_names (List[str], optional): [description]. Defaults to None.
        class_types (List[str], optional): 'event' or 'segment'. Defaults to None.
        store_type ([type], optional): [description]. Defaults to zarr.TemporaryStore.
        store_name (str, optional): [description]. Defaults to 'store.zarr'.
        chunk_len (int, optional): [description]. Defaults to 1_000_000.

    Raises:
        ValueError: [description]
        ValueError: [description]

    Returns:
        [type]: [description]
    """

    if class_names is not None and nb_classes is not None and len(class_names) != nb_classes:
        raise ValueError(f"Number of classes ({nb_classes}) needs to match len(class_names) ({len(class_names)}).")
    if class_types is not None and nb_classes is not None and len(class_names) != nb_classes:
        raise ValueError(f"Number of classes ({nb_classes}) needs to match len(class_names) ({len(class_types)}).")

    # initialize the store
    store = store_type(store_name)
    root = zarr.group(store=store, overwrite=True)  # need to def the root
    for target in ["train", "val", "test"]:
        root.empty(name=f"{target}/x", shape=(0, nb_channels), chunks=(chunk_len, nb_channels), dtype=np.float16)
        root.empty(name=f"{target}/y", shape=(0, nb_classes), chunks=(chunk_len, nb_classes), dtype=np.float16)
        # root.empty(name=f'{target}/eventtimes', shape=(0, nb_classes), chunks=(1_000,), dtype=np.float)
        if make_single_class_datasets:
            for class_name in class_names[1:]:
                root.empty(name=f"{target}/y_{class_name}", shape=(0, 2), chunks=(chunk_len, nb_classes), dtype=np.float16)

    # init metadata - since attrs cannot be appended to, we init a dict here, populate it with information below and finaly assign it to root.attrs
    root.attrs["samplerate_x_Hz"] = samplerate
    root.attrs["samplerate_y_Hz"] = samplerate

    root.attrs["class_names"] = [str(cn) for cn in class_names]
    root.attrs["class_types"] = class_types

    if make_single_class_datasets:
        for class_name, class_type in zip(class_names[1:], class_types[1:]):
            root.attrs[f"class_names_{class_name}"] = [class_names[0], class_name]
            root.attrs[f"class_types_{class_name}"] = [class_types[0], class_type]

    for target in ["train", "val", "test"]:
        root.attrs[f"filename_startsample_{target}"] = []
        root.attrs[f"filename_endsample_{target}"] = []
        root.attrs[f"filename_{target}"] = []
    return root


def events_to_probabilities(eventsamples: List[int], desired_len: Optional[int] = None, extent: int = 61):
    """Converts list of events to one-hot-encoded probability vectors.

    Args:
        eventsamples (List[int]): List of event "times" in samples.
        desired_len (float, optional): Length of the probability vector.
                                       Events exceeding `desired_len` will be ignored.
                                       Defaults to `max(eventsamples) + extent`.
        extent (int, optional): Temporal extent of an event in the probability vector.
                                Each event will be represented as a box with a duration `exent` samples centered on the event.
                                Defaults to 61 samples (+/-30 samples).
    Returns:
        probabilities: np.array with shape [desired_len, 2]
                       where `probabilities[:, 0]` corresponds to the probability of no event
                       and `probabilities[:, 0]` corresponds to the probability of an event.
    """
    if desired_len is None:
        desired_len = max(eventsamples) + extent
    else:
        eventsamples = eventsamples[eventsamples < desired_len - extent]  # delete all eventsamples exceeding desired_len
    probabilities = np.zeros((desired_len, 2))
    probabilities[eventsamples, 1] = 1
    probabilities[:, 1] = np.convolve(probabilities[:, 1], np.ones((extent,)), mode="same")
    probabilities[:, 0] = 1 - probabilities[:, 1]
    return probabilities


def infer_class_info(df: pd.DataFrame):
    """[summary]

    Args:
        df ([type]): [description]

    Returns:
        [type]: [description]
    """
    class_names, first_indices = np.unique(df["name"], return_index=True)
    class_names = list(class_names)
    class_names.insert(0, "noise")

    # infer class type - event if start and end are the same
    class_types = ["segment"]
    for first_index in first_indices:
        if df.loc[first_index]["start_seconds"] == df.loc[first_index]["stop_seconds"]:
            class_types.append("event")
        else:
            class_types.append("segment")
    return class_names, class_types


def make_annotation_matrix(
    df: pd.DataFrame, nb_samples: int, samplerate: float, class_names: Optional[List[str]] = None
) -> np.ndarray:
    """One-hot encode a list of song timings to a binary matrix.

    Args:
        df (pd.DataFrame): DataFrame with the following columns:
                            - name: class name of the syllable/song event
                            - start_seconds: start of the song event in the audio recording in seconds.
                            - stop_seconds: stop of the song event in the audio recording in seconds.
        nb_samples ([type]): Length of the annotation matrix in samples.
        samplerate (float): Sample rate for the annotation matrix in Hz.
        class_names (List[str], optional): List of class names.
                            If provided, the annotation matrix will be built only for the events in class_names.
                            Otherwise, the matrix will be build for all class names in the df.
                            Order in class_names determines order in class_matrix

    Returns:
        nd.array: Binary matrix [nb_samples, nb_classes]
                  with 1 indicating the presence of a class at a specific sample.
    """
    if class_names is None:
        class_names, _ = infer_class_info(df)
    class_matrix = np.zeros((nb_samples, len(class_names)))
    for _, row in df.iterrows():
        if not row["name"] in class_names:
            continue
        if np.all(np.isnan(row["start_seconds"])):
            continue
        class_index = class_names.index(row["name"])
        start_index = int(row["start_seconds"] * samplerate)
        stop_index = int(row["stop_seconds"] * samplerate + 1)
        if start_index < stop_index:
            class_matrix[start_index:stop_index, class_index] = 1
        else:
            logger.warning(f"{start_index} should be greater than {stop_index} for row {row}")
    return class_matrix


def normalize_probabilities(p: np.ndarray) -> np.ndarray:
    """[summary]

    Args:
        p (np.ndarray): [description]

    Returns:
        np.ndarray: [description]
    """
    p_song = np.sum(p[:, 1:], axis=-1)

    p[p_song > 1.0, 1:] = p[p_song > 1.0, 1:] / p_song[p_song > 1.0, np.newaxis]
    p[:, 0] = 1 - np.sum(p[:, 1:], axis=-1)
    return p


def make_gaps(
    y: np.ndarray,
    gap_seconds: float,
    samplerate: float,
    start_seconds: Optional[List[float]] = None,
    stop_seconds: Optional[List[float]] = None,
) -> np.ndarray:
    """[summary]

    0011112222000111100 -> 0011100222000111100 (gap_fullwidth=2)

    Args:
        y (np.ndarray): One-hot encoded labels [T, nb_labels]
        gap_seconds (float): [description]
        samplerate (float): [description]
        start_seconds:
        stop_seconds:

    Returns:
        np.ndarray: [description]
    """
    y0 = y.copy()

    if y.ndim > 1 and y.shape[1] > 1:
        y = np.argmax(y, axis=1)

    gap_halfwidth = int(np.floor(gap_seconds * samplerate) / 2)

    # widen gaps between adjacent syllables of different types
    a = y.copy().astype(np.float)
    label_change = np.where(np.diff(a, axis=0) != 0)[0]
    # remove on and offsets (0->label or label->0)
    onset = a[label_change] == 0
    offset = a[label_change + 1] == 0
    neither_on_nor_off = np.logical_and(~onset, ~offset)

    if np.sum(neither_on_nor_off):
        label_change = label_change[neither_on_nor_off]

        # introduce gap around label changes for adjacent syllables
        for gap_offset in range(-gap_halfwidth, gap_halfwidth + 1):
            y[label_change + gap_offset] = 0

    # one-hot-encode gapped labels
    y0[:] = 0
    for label in range(y0.shape[1]):
        y0[y == label, label] = 1

    # widen gaps between syllables of same type
    for label in range(1, y0.shape[1]):
        label_change = np.where(np.diff(y0[:, label], axis=0) != 0)[0]
        onset = y0[label_change, label] == 0
        offset = y0[label_change + 1, label] == 0

        # there is no gap before the first syll starts and after the last syll ends so ignore those
        gap_onsets = label_change[onset][1:]
        gap_offsets = label_change[offset][:-1]

        # just to be safe - remove all offsets occurring before the first onset and all onsets occurring before the last offset here
        if len(gap_offsets) > 0 and len(gap_onsets) > 0:
            gap_offsets = gap_offsets[gap_offsets > np.min(gap_onsets)]
        # need to check twice since len(gap_offsets) might change above
        if len(gap_offsets) > 0 and len(gap_onsets) > 0:
            gap_onsets = gap_onsets[gap_onsets < np.max(gap_offsets)]

        if len(gap_offsets) > 0 and len(gap_onsets) > 0:
            gaps = gap_onsets - gap_offsets
        else:
            gaps = []
            gap_onsets = []
            gap_offsets = []

        for gap, gap_onset, gap_offset in zip(gaps, gap_onsets, gap_offsets):
            if gap < 2 * gap_halfwidth:
                midpoint = int(gap_offset + gap / 2)
                y0[midpoint - gap_halfwidth : midpoint + gap_halfwidth + 1, :] = 0

    # ensure gaps exist even when same-type segments touch
    if start_seconds is not None and stop_seconds is not None:
        start_samples = np.array(start_seconds * samplerate).astype(np.uintp)
        stop_samples = np.array(stop_seconds * samplerate).astype(np.uintp)
        for start_sample, stop_sample in zip(start_samples, stop_samples):
            y0[start_sample : int(start_sample + gap_halfwidth), :] = 0
            y0[int(stop_sample - gap_halfwidth) : stop_sample, :] = 0

    return y0


def blur_events(event_trace: np.ndarray, event_std_seconds: float, samplerate: float) -> np.ndarray:
    """Blur event trace with a gaussian.

    Args:
        event_trace (np.ndarray): shape (N,)
        event_std_seconds (float): With of the Gaussian in seconds
        samplerate (float): sample rate of event_trace

    Returns:
        np.ndarray: blurred event trace
    """
    event_std_samples = event_std_seconds * samplerate
    win = scipy.signal.gaussian(int(event_std_samples * 8), std=event_std_samples)
    event_trace = scipy.signal.convolve(event_trace.astype(float), win, mode="same")
    return event_trace



import hashlib
from pathlib import Path
from typing import Union


def hash_data(data_path: Union[str, Path], chunk_size: int = 65536) -> str:
    """Compute MD5 hash of the data_path (dir or file) for data versioning.

    Args:
        data_path ([type]): [description]
        chunk_size (int, optional): [description]. Defaults to 65536.

    Raises:
        ValueError: [description]

    Returns:
        str: hexadecimal string containing the hash code
    """
    if Path(data_path).is_dir():
        hash = _hash_dir(data_path, chunk_size)
    elif Path(data_path).is_file():
        hash = _hash_file(data_path, chunk_size)
    else:
        raise ValueError(f"{data_path} is neither directory nor file.")
    return hash.hexdigest()


def _update_hash_dir(directory: Union[str, Path], hash, chunk_size: int):
    # from https://stackoverflow.com/questions/24937495/how-can-i-calculate-a-hash-for-a-filesystem-directory-using-python/54477583#54477583
    assert Path(directory).is_dir()
    for path in sorted(Path(directory).iterdir(), key=lambda p: str(p).lower()):
        hash.update(path.name.encode())
        if path.is_file():
            with open(path, "rb") as f:
                for chunk in iter(lambda: f.read(chunk_size), b""):
                    hash.update(chunk)
        elif path.is_dir():
            hash = _update_hash_dir(path, hash, chunk_size)
    return hash


def _hash_dir(directory: Union[str, Path], chunk_size: int):
    return _update_hash_dir(directory, hashlib.md5(), chunk_size)


def _hash_file(data_file: Union[str, Path], chunk_size: int):
    hash = hashlib.md5()
    with open(data_file, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            hash.update(chunk)
    return hash



from setuptools import setup, find_packages
import codecs
import re
import os

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='das',
    version=find_version("src/das/__init__.py"),
)


"""Defines the network architectures."""
import tensorflow.keras as keras
import tensorflow.keras.layers as kl
from typing import List, Optional
from . import tcn as tcn_layer
from .kapre.time_frequency import Spectrogram
from .kapre.utils import AmplitudeToDB


model_dict = dict()


def _register_as_model(func):
    """Adds func to model_dict Dict[modelname: modelfunc]. For selecting models by string."""
    model_dict[func.__name__] = func
    return func


@_register_as_model
def cnn(
    nb_freq,
    nb_classes,
    nb_channels=1,
    nb_hist=1,
    nb_filters=16,
    nb_stacks=2,
    kernel_size=3,
    nb_conv=3,
    loss="categorical_crossentropy",
    batch_norm=False,
    return_sequences=False,
    sample_weight_mode: str = None,
    learning_rate: float = 0.0001,
    **kwignored
):
    """CNN for single-frequency and multi-channel data - uses 1D convolutions.

    Args:
        nb_freq ([type]): [description]
        nb_classes ([type]): [description]
        nb_channels (int, optional): [description]. Defaults to 1.
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        nb_stacks (int, optional): [description]. Defaults to 2.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 3.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        batch_norm (bool, optional): [description]. Defaults to False.
        return_sequences (bool, optional): [description]. Defaults to False.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred

    Raises:
        ValueError: When trying to init model with return_sequences.
                    The cnn only does instance classificaton, not regression (sequence prediction).

    Returns:
        [type]: [description]
    """
    if return_sequences:
        raise ValueError("Cannot perform regression with CNN.")
    inp = kl.Input(shape=(nb_hist, nb_channels))
    out = inp
    for conv in range(nb_conv):
        for _ in range(nb_stacks):
            out = kl.Conv1D(nb_filters * (2 ** conv), kernel_size, padding="same", activation="relu")(out)
            out = kl.BatchNormalization()(out) if batch_norm else out
        out = kl.MaxPooling1D(min(int(out.shape[1]), 2))(out)

    out = kl.Flatten()(out)
    out = kl.Dense(nb_classes * 4, activation="relu")(out)
    out = kl.Dense(nb_classes * 2, activation="relu")(out)
    out = kl.Dense(nb_classes, activation="relu")(out)
    out = kl.Activation("softmax")(out)

    model = keras.models.Model(inp, out, name="FCN")
    model.compile(
        optimizer=keras.optimizers.Adam(lr=learning_rate, amsgrad=True), loss=loss, sample_weight_mode=sample_weight_mode
    )
    return model


@_register_as_model
def cnn2D(
    nb_freq,
    nb_classes,
    nb_channels=1,
    nb_hist=1,
    nb_filters=16,
    nb_stacks=2,
    kernel_size=3,
    nb_conv=3,
    loss="categorical_crossentropy",
    batch_norm=False,
    return_sequences=False,
    sample_weight_mode: str = None,
    learning_rate: float = 0.0001,
    **kwignored
):
    """CNN for multi-frequency and multi-channel data - uses 2D convolutions.

    Args:
        nb_freq ([type]): [description]
        nb_classes ([type]): [description]
        nb_channels (int, optional): [description]. Defaults to 1.
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        nb_stacks (int, optional): [description]. Defaults to 2.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 3.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        batch_norm (bool, optional): [description]. Defaults to False.
        return_sequences (bool, optional): [description]. Defaults to False.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred

    Raises:
        ValueError: When trying to init model with return_sequences.
                    The cnn only does instance classificaton, not regression (sequence prediction).

    Returns:
        [type]: [description]
    """
    inp = kl.Input(shape=(nb_hist, nb_freq, nb_channels))
    out = inp
    for conv in range(nb_conv):
        for stack in range(nb_stacks):
            out = kl.Conv2D(
                nb_filters * (2 ** conv), (max(1, int(out.shape[1])), kernel_size), padding="same", activation="relu"
            )(out)
            out = kl.BatchNormalization()(out) if batch_norm else out

        out = kl.MaxPooling2D((min(int(out.shape[1]), 2), 2))(out)
    out = kl.Flatten()(out)
    out = kl.Dense(int(nb_classes * 8 * nb_filters), activation="relu")(out)
    out = kl.Dropout(0.1)(out)
    out = kl.Dense(int(nb_classes * 4 * nb_filters), activation="relu")(out)
    out = kl.Dropout(0.1)(out)
    out = kl.Dense(int(nb_classes * 2 * nb_filters), activation="relu")(out)
    out = kl.Dropout(0.1)(out)
    out = kl.Dense(nb_classes, activation="relu")(out)
    out = kl.Activation("softmax")(out)

    model = keras.models.Model(inp, out, name="CNN2D")
    model.compile(
        optimizer=keras.optimizers.Adam(lr=learning_rate, amsgrad=True), loss=loss, sample_weight_mode=sample_weight_mode
    )
    return model


@_register_as_model
def fcn(
    nb_freq,
    nb_classes,
    nb_channels=1,
    nb_hist=1,
    nb_filters=16,
    nb_stacks=2,
    kernel_size=3,
    nb_conv=3,
    loss="categorical_crossentropy",
    batch_norm=False,
    return_sequences=True,
    sample_weight_mode: str = None,
    learning_rate: float = 0.0001,
    **kwignored
):
    """[summary]

    Args:
        nb_freq ([type]): [description]
        nb_classes ([type]): [description]
        nb_channels (int, optional): [description]. Defaults to 1.
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        nb_stacks (int, optional): [description]. Defaults to 2.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 3.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        batch_norm (bool, optional): [description]. Defaults to False.
        return_sequences (bool, optional): [description]. Defaults to True.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred

    Returns:
        [type]: [description]
    """
    inp = kl.Input(shape=(nb_hist, nb_channels))
    out = inp
    for conv in range(nb_conv):
        for _ in range(nb_stacks):
            out = kl.Conv1D(nb_filters * (2 ** conv), kernel_size, padding="same", activation="relu")(out)
            out = kl.BatchNormalization()(out) if batch_norm else out
        out = kl.MaxPooling1D(min(int(out.shape[1]), 2))(out)

    for conv in range(nb_conv, 0, -1):
        out = kl.UpSampling1D(size=2)(out)
        out = kl.Conv1D(nb_filters * (2 ** conv), kernel_size, padding="same", activation="relu")(out)

    if not return_sequences:
        out = kl.Flatten()(out)
    out = kl.Dense(nb_classes)(out)
    out = kl.Activation("softmax")(out)

    model = keras.models.Model(inp, out, name="FCN")
    model.compile(
        optimizer=keras.optimizers.Adam(lr=learning_rate, amsgrad=True), loss=loss, sample_weight_mode=sample_weight_mode
    )

    return model


@_register_as_model
def fcn2D(
    nb_freq,
    nb_classes,
    nb_channels=1,
    nb_hist=1,
    nb_filters=16,
    nb_stacks=2,
    kernel_size=3,
    nb_conv=3,
    loss="categorical_crossentropy",
    batch_norm=False,
    return_sequences=True,
    sample_weight_mode: str = None,
    learning_rate: float = 0.0005,
    **kwignored
):
    """[summary]

    Args:
        nb_freq ([type]): [description]
        nb_classes ([type]): [description]
        nb_channels (int, optional): [description]. Defaults to 1.
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        nb_stacks (int, optional): [description]. Defaults to 2.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 3.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        batch_norm (bool, optional): [description]. Defaults to False.
        return_sequences (bool, optional): [description]. Defaults to True.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred

    Returns:
        [type]: [description]
    """
    inp = kl.Input(shape=(nb_hist, nb_freq, nb_channels))
    out = inp
    for conv in range(nb_conv):
        for stack in range(nb_stacks):
            out = kl.Conv2D(
                nb_filters * (2 ** conv), (max(1, int(out.shape[1])), kernel_size), padding="same", activation="relu"
            )(out)
            out = kl.BatchNormalization()(out) if batch_norm else out
        out = kl.MaxPooling2D((min(int(out.shape[1]), 2), 2))(out)
    for conv in range(nb_conv, 0, -1):
        out = kl.Conv2D(
            int(nb_filters * (conv / 2)), (max(1, int(out.shape[1])), kernel_size), padding="same", activation="relu"
        )(out)
        out = kl.BatchNormalization()(out) if batch_norm else out

    out = kl.Flatten()(out)
    out = kl.Dense(nb_classes, activation="relu")(out)
    out = kl.Activation("softmax")(out)

    model = keras.models.Model(inp, out, name="CNN")
    model.compile(optimizer=keras.optimizers.Adam(lr=learning_rate, amsgrad=True), loss=loss)
    return model


@_register_as_model
def tcn_seq(
    nb_freq: int,
    nb_classes: int,
    nb_hist: int = 1,
    nb_filters: int = 16,
    kernel_size: int = 3,
    nb_conv: int = 1,
    loss: str = "categorical_crossentropy",
    dilations: List[int] = [1, 2, 4, 8, 16],
    activation: str = "norm_relu",
    use_skip_connections: bool = True,
    return_sequences: bool = True,
    dropout_rate: float = 0.00,
    padding: str = "same",
    sample_weight_mode: str = None,
    nb_pre_conv: int = 0,
    learning_rate: float = 0.0001,
    upsample: bool = True,
    out_activation: str = "softmax",
    use_separable: bool = False,
    **kwignored
):
    """Create TCN network.

    Args:
        nb_freq (int): [description]
        nb_classes (int): [description]
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 1.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        dilations (List[int], optional): [description]. Defaults to [1, 2, 4, 8, 16].
        activation (str, optional): [description]. Defaults to 'norm_relu'.
        use_skip_connections (bool, optional): [description]. Defaults to True.
        return_sequences (bool, optional): [description]. Defaults to True.
        dropout_rate (float, optional): [description]. Defaults to 0.00.
        padding (str, optional): [description]. Defaults to 'same'.
        nb_pre_conv (int, optional): number of conv-relu-batchnorm-maxpool2 blocks before the TCN - useful for reducing the sample rate. Defaults to 0
        upsample (bool, optional): whether or not to restore the model output to the input samplerate.
                                   Should generally be True during training and evaluation but my speed up inference .
                                   Defaults to True.
        out_activation (str, optional): activation type for the output. Defaults to 'softmax'.
        use_separable (bool, optional): use separable convs in residual block. Defaults to False.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ignored

    Returns:
        [keras.models.Model]: Compiled TCN network model.
    """
    input_layer = kl.Input(shape=(nb_hist, nb_freq))
    out = input_layer
    for conv in range(nb_pre_conv):
        out = kl.Conv1D(nb_filters * (2 ** conv), kernel_size, padding="same", activation="relu")(out)
        out = kl.BatchNormalization()(out)
        out = kl.MaxPooling1D(min(int(out.shape[1]), 2))(out)

    x = tcn_layer.TCN(
        nb_filters=nb_filters,
        kernel_size=kernel_size,
        nb_stacks=nb_conv,
        dilations=dilations,
        activation=activation,
        use_skip_connections=use_skip_connections,
        padding=padding,
        dropout_rate=dropout_rate,
        return_sequences=return_sequences,
        use_separable=use_separable,
    )(out)
    x = kl.Dense(nb_classes)(x)
    x = kl.Activation(out_activation)(x)
    if nb_pre_conv > 0 and upsample:
        x = kl.UpSampling1D(size=2 ** nb_pre_conv)(x)
    output_layer = x

    model = keras.models.Model(input_layer, output_layer, name="TCN")
    model.compile(
        optimizer=keras.optimizers.Adam(lr=learning_rate, amsgrad=True, clipnorm=1.0),
        loss=loss,
        sample_weight_mode=sample_weight_mode,
    )
    return model


@_register_as_model
def tcn(*args, **kwargs):
    """Synonym for tcn_stft."""
    return tcn_stft(*args, **kwargs)


@_register_as_model
def tcn_tcn(
    nb_freq: int,
    nb_classes: int,
    nb_hist: int = 1,
    nb_filters: int = 16,
    kernel_size: int = 3,
    nb_conv: int = 1,
    loss: str = "categorical_crossentropy",
    dilations: List[int] = [1, 2, 4, 8, 16],
    activation: str = "norm_relu",
    use_skip_connections: bool = True,
    return_sequences: bool = True,
    dropout_rate: float = 0.00,
    padding: str = "same",
    sample_weight_mode: str = None,
    nb_pre_conv: int = 0,
    learning_rate: float = 0.0005,
    upsample: bool = True,
    use_separable: bool = False,
    **kwignored
):
    """Create TCN network with TCN layer as pre-processing and downsampling frontend.

    Args:
        nb_freq (int): [description]
        nb_classes (int): [description]
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 1.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        dilations (List[int], optional): [description]. Defaults to [1, 2, 4, 8, 16].
        activation (str, optional): [description]. Defaults to 'norm_relu'.
        use_skip_connections (bool, optional): [description]. Defaults to True.
        return_sequences (bool, optional): [description]. Defaults to True.
        dropout_rate (float, optional): [description]. Defaults to 0.00.
        padding (str, optional): [description]. Defaults to 'same'.
        nb_pre_conv (int, optional): If >0 adds a single TCN layer with a final maxpooling layer
                                     with block size of `2**nb_pre_conv` before the TCN.
                                     Useful for speeding up training by reducing the sample rate early in the network.
                                     Defaults to 0 (no downsampling)
        learning_rate (float, optional) Defaults to 0.0005
        upsample (bool, optional): whether or not to restore the model output to the input samplerate.
                                   Should generally be True during training and evaluation but my speed up inference .
                                   Defaults to True.
        use_separable (bool, optional): use separable convs in residual block. Defaults to False.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred

    Returns:
        [keras.models.Model]: Compiled TCN network model.
    """
    input_layer = kl.Input(shape=(nb_hist, nb_freq))
    out = input_layer
    if nb_pre_conv > 0:
        out = tcn_layer.TCN(
            nb_filters=nb_filters,
            kernel_size=kernel_size,
            nb_stacks=nb_pre_conv,
            dilations=dilations,
            activation=activation,
            use_skip_connections=use_skip_connections,
            padding=padding,
            dropout_rate=dropout_rate,
            return_sequences=return_sequences,
            use_separable=use_separable,
            name="frontend",
        )(out)
        out = kl.MaxPooling1D(pool_size=2 ** nb_pre_conv, strides=2 ** nb_pre_conv)(out)  # or avg pooling?

    x = tcn_layer.TCN(
        nb_filters=nb_filters,
        kernel_size=kernel_size,
        nb_stacks=nb_conv,
        dilations=dilations,
        activation=activation,
        use_skip_connections=use_skip_connections,
        padding=padding,
        dropout_rate=dropout_rate,
        return_sequences=return_sequences,
        use_separable=use_separable,
    )(out)
    x = kl.Dense(nb_classes)(x)
    x = kl.Activation("softmax")(x)
    if nb_pre_conv > 0 and upsample:
        x = kl.UpSampling1D(size=2 ** nb_pre_conv)(x)
    output_layer = x
    model = keras.models.Model(input_layer, output_layer, name="TCN")

    model.compile(
        optimizer=keras.optimizers.Adam(lr=learning_rate, amsgrad=True, clipnorm=1.0),
        loss=loss,
        sample_weight_mode=sample_weight_mode,
    )
    return model


@_register_as_model
def tcn_small(
    nb_freq: int,
    nb_classes: int,
    nb_hist: int = 1,
    nb_filters: int = 16,
    kernel_size: int = 3,
    nb_conv: int = 1,
    loss: str = "categorical_crossentropy",
    dilations: List[int] = [1, 2, 4, 8, 16],
    activation: str = "norm_relu",
    use_skip_connections: bool = True,
    return_sequences: bool = True,
    dropout_rate: float = 0.00,
    padding: str = "same",
    sample_weight_mode: str = None,
    nb_pre_conv: int = 0,
    learning_rate: float = 0.0005,
    upsample: bool = True,
    use_separable: bool = False,
    **kwignored
):
    """Create TCN network with TCN layer as pre-processing and downsampling frontend.

    Args:
        nb_freq (int): [description]
        nb_classes (int): [description]
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 1.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        dilations (List[int], optional): [description]. Defaults to [1, 2, 4, 8, 16].
        activation (str, optional): [description]. Defaults to 'norm_relu'.
        use_skip_connections (bool, optional): [description]. Defaults to True.
        return_sequences (bool, optional): [description]. Defaults to True.
        dropout_rate (float, optional): [description]. Defaults to 0.00.
        padding (str, optional): [description]. Defaults to 'same'.
        nb_pre_conv (int, optional): If >0 adds a single TCN layer with a final maxpooling layer
                                     with block size of `2**nb_pre_conv` before the TCN.
                                     Useful for speeding up training by reducing the sample rate early in the network.
                                     Defaults to 0 (no downsampling)
        learning_rate (float, optional) Defaults to 0.0005
        upsample (bool, optional): whether or not to restore the model output to the input samplerate.
                                   Should generally be True during training and evaluation but my speed up inference .
                                   Defaults to True.
        use_separable (bool, optional): use separable convs in residual block. Defaults to False.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred

    Returns:
        [keras.models.Model]: Compiled TCN network model.
    """
    input_layer = kl.Input(shape=(nb_hist, nb_freq))
    out = input_layer
    if nb_pre_conv > 0:
        out = tcn_layer.TCN(
            nb_filters=32,
            kernel_size=3,
            nb_stacks=1,
            dilations=dilations,
            activation=activation,
            use_skip_connections=use_skip_connections,
            padding=padding,
            dropout_rate=dropout_rate,
            return_sequences=return_sequences,
            use_separable=use_separable,
            name="frontend",
        )(out)
        out = kl.MaxPooling1D(pool_size=2 ** nb_pre_conv, strides=2 ** nb_pre_conv)(out)  # or avg pooling?

    x = tcn_layer.TCN(
        nb_filters=nb_filters,
        kernel_size=kernel_size,
        nb_stacks=nb_conv,
        dilations=dilations,
        activation=activation,
        use_skip_connections=use_skip_connections,
        padding=padding,
        dropout_rate=dropout_rate,
        return_sequences=return_sequences,
        use_separable=use_separable,
    )(out)
    x = kl.Dense(nb_classes)(x)
    x = kl.Activation("softmax")(x)
    if nb_pre_conv > 0 and upsample:
        x = kl.UpSampling1D(size=2 ** nb_pre_conv)(x)
    output_layer = x

    model = keras.models.Model(input_layer, output_layer, name="TCN")
    model.compile(
        optimizer=keras.optimizers.Adam(lr=learning_rate, amsgrad=True, clipnorm=1.0),
        loss=loss,
        sample_weight_mode=sample_weight_mode,
    )
    return model


@_register_as_model
def tcn_stft(
    nb_freq: int,
    nb_classes: int,
    nb_hist: int = 1,
    nb_filters: int = 16,
    kernel_size: int = 3,
    nb_conv: int = 1,
    loss: str = "categorical_crossentropy",
    dilations: Optional[List[int]] = None,
    activation: str = "norm_relu",
    use_skip_connections: bool = True,
    return_sequences: bool = True,
    dropout_rate: float = 0.00,
    padding: str = "same",
    sample_weight_mode: str = None,
    nb_pre_conv: int = 0,
    pre_nb_dft: int = 64,
    nb_lstm_units: int = 0,
    learning_rate: float = 0.0005,
    upsample: bool = True,
    use_separable: bool = False,
    **kwignored
):
    """Create TCN network with optional trainable STFT layer as pre-processing and downsampling frontend.

    Args:
        nb_freq (int): [description]
        nb_classes (int): [description]
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 1.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        dilations (List[int], optional): [description]. Defaults to [1, 2, 4, 8, 16].
        activation (str, optional): [description]. Defaults to 'norm_relu'.
        use_skip_connections (bool, optional): [description]. Defaults to True.
        return_sequences (bool, optional): [description]. Defaults to True.
        dropout_rate (float, optional): [description]. Defaults to 0.00.
        padding (str, optional): [description]. Defaults to 'same'.
        nb_pre_conv (int, optional): If >0 adds a single STFT layer with a hop size of 2**nb_pre_conv before the TCN.
                                     Useful for speeding up training by reducing the sample rate early in the network.
                                     Defaults to 0 (no downsampling)
        pre_nb_dft (int, optional): Number of filters (roughly corresponding to filters) in the STFT frontend.
                                    Defaults to 64.
        learning_rate (float, optional) Defaults to 0.0005
        nb_lstm_units (int, optional): Defaults to 0.
        upsample (bool, optional): whether or not to restore the model output to the input samplerate.
                                   Should generally be True during training and evaluation but may speed up inference.
                                   Defaults to True.
        use_separable (bool, optional): use separable convs in residual block. Defaults to False.
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred

    Returns:
        [keras.models.Model]: Compiled TCN network model.
    """
    if dilations is None:
        dilations = [1, 2, 4, 8, 16]
    # if nb_freq > 1:
    #     raise ValueError(f'This model only works with single channel data but last dim of inputs has len {nb_freq} (should be 1).')
    input_layer = kl.Input(shape=(nb_hist, nb_freq))
    out = input_layer
    if nb_pre_conv > 0:
        out = Spectrogram(
            n_dft=pre_nb_dft,
            n_hop=2 ** nb_pre_conv,
            return_decibel_spectrogram=True,
            power_spectrogram=1.0,
            trainable_kernel=True,
            name="trainable_stft",
            image_data_format="channels_last",
        )(out)
        # out = AmplitudeToDB()(out)
        out = kl.Reshape((out.shape[1], out.shape[2] * out.shape[3]))(out)

    x = tcn_layer.TCN(
        nb_filters=nb_filters,
        kernel_size=kernel_size,
        nb_stacks=nb_conv,
        dilations=dilations,
        activation=activation,
        use_skip_connections=use_skip_connections,
        padding=padding,
        dropout_rate=dropout_rate,
        return_sequences=return_sequences,
        use_separable=use_separable,
    )(out)

    if nb_lstm_units > 0:
        x = kl.Bidirectional(kl.LSTM(units=nb_lstm_units, return_sequences=True))(x)

    x = kl.Dense(nb_classes)(x)
    x = kl.Activation("softmax")(x)

    if nb_pre_conv > 0 and upsample:
        x = kl.UpSampling1D(size=2 ** nb_pre_conv)(x)

    output_layer = x

    model = keras.models.Model(input_layer, output_layer, name="TCN")
    model.compile(
        optimizer=keras.optimizers.Adam(lr=learning_rate, amsgrad=True, clipnorm=1.0),
        loss=loss,
        sample_weight_mode=sample_weight_mode,
    )
    return model


@_register_as_model
def tcn_multi(
    nb_freq: int,
    nb_classes: int,
    nb_hist: int = 1,
    nb_filters: int = 16,
    kernel_size: int = 3,
    nb_conv: int = 1,
    loss: str = "categorical_crossentropy",
    dilations: List[int] = [1, 2, 4, 8, 16],
    activation: str = "norm_relu",
    use_skip_connections: bool = True,
    return_sequences: bool = True,
    dropout_rate: float = 0.00,
    padding: str = "same",
    sample_weight_mode: str = None,
    learning_rate: float = 0.0005,
    use_separable: bool = False,
    pre_kernel_size: int = 16,
    pre_nb_filters: int = 16,
    pre_nb_conv: int = 2,
    **kwignored
):
    """Create TCN network with TCN layer as pre-processing and downsampling frontend with weights shared between channels.

    Args:
        nb_freq (int): [description]
        nb_classes (int): [description]
        nb_hist (int, optional): [description]. Defaults to 1.
        nb_filters (int, optional): [description]. Defaults to 16.
        kernel_size (int, optional): [description]. Defaults to 3.
        nb_conv (int, optional): [description]. Defaults to 1.
        loss (str, optional): [description]. Defaults to "categorical_crossentropy".
        dilations (List[int], optional): [description]. Defaults to [1, 2, 4, 8, 16].
        activation (str, optional): [description]. Defaults to 'norm_relu'.
        use_skip_connections (bool, optional): [description]. Defaults to True.
        return_sequences (bool, optional): [description]. Defaults to True.
        dropout_rate (float, optional): [description]. Defaults to 0.00.
        padding (str, optional): [description]. Defaults to 'same'.
        learning_rate (float, optional) Defaults to 0.0005
        use_separable (bool, optional): use separable convs in residual block. Defaults to False.
        nb_pre_conv (int, optional): If >0 adds a single STFT layer with a hop size of 2**nb_pre_conv before the TCN.
                                     Useful for speeding up training by reducing the sample rate early in the network.
                                     Defaults to 0 (no downsampling)
        pre_nb_filters, pre_kernelsize, pre_nb_conv
        kwignored (Dict, optional): additional kw args in the param dict used for calling m(**params) to be ingonred


    Returns:
        [keras.models.Model]: Compiled TCN network model.
    """

    # define the per-channel model
    nb_channels = nb_freq
    channels_in = []
    for chan in range(nb_channels):
        channels_in.append(kl.Input(shape=(nb_hist, 1), name="channel_{0}".format(chan)))

    # channel model will be shared, weights and all
    channel_model = tcn_layer.TCN_new(
        nb_filters=pre_nb_filters,
        kernel_size=pre_kernel_size,
        nb_stacks=pre_nb_conv,
        dilations=dilations,
        activation="relu",
        use_skip_connections=use_skip_connections,
        padding=padding,
        dropout_rate=dropout_rate,
        return_sequences=return_sequences,
        use_separable=use_separable,
    )  # , name='channel')
    # if nb_pre_conv > 0:
    #     out = tcn_layer.TCN(nb_filters=32, kernel_size=3, nb_stacks=1, dilations=dilations,
    #                         activation=activation, use_skip_connections=use_skip_connections, padding=padding,
    #                         dropout_rate=dropout_rate, return_sequences=return_sequences,
    #                         use_separable=use_separable, name='frontend')(out)
    #     out = kl.MaxPooling1D(pool_size=2**nb_pre_conv, strides=2**nb_pre_conv)(out)  # or avg pooling?
    channels_out = []
    for chan in channels_in:
        channels_out.append(channel_model(chan))

    out = kl.concatenate(channels_out)

    x = tcn_layer.TCN(
        nb_filters=nb_filters,
        kernel_size=kernel_size,
        nb_stacks=nb_conv,
        dilations=dilations,
        activation=activation,
        use_skip_connections=use_skip_connections,
        padding=padding,
        dropout_rate=dropout_rate,
        return_sequences=return_sequences,
        name="merge",
        use_separable=use_separable,
    )(out)

    x = kl.Dense(nb_classes)(x)
    x = kl.Activation("softmax")(x)
    output_layer = x

    model = keras.models.Model(channels_in, output_layer, name="TCN")
    model.compile(
        optimizer=keras.optimizers.Adam(lr=learning_rate, amsgrad=True, clipnorm=1.0),
        loss=loss,
        sample_weight_mode=sample_weight_mode,
    )
    return model



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import numpy as np
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer
from . import backend
from . import backend_keras
from typing import Optional


class AmplitudeToDB(Layer):
    """Converts spectrogram values to decibels.

    Examples:
        Adding dB conversion after a spectrogram:

        >>> model.add(Spectrogram(return_decibel=False))
        >>> model.add(AmplitudeToDB())
        which is the same as:
        >>> model.add(Spectrogram(return_decibel=True))

    """

    def __init__(self, amin: float = 1e-10, top_db: float = 80.0, **kwargs):
        """Args:
        amin (float, optional): Noise floor. Defaults to 1e-10 (dB).
        top_db (float, optional): Dynamic range of output. Defaults to 80.0 (dB).
        """
        self.amin = amin
        self.top_db = top_db
        super(AmplitudeToDB, self).__init__(**kwargs)

    def call(self, x, mask=None):
        return backend_keras.amplitude_to_decibel(x, amin=self.amin, dynamic_range=self.top_db)

    def get_config(self):
        config = {"amin": self.amin, "top_db": self.top_db}
        base_config = super(AmplitudeToDB, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class Normalization2D(Layer):
    """Normalizes input along an axis.

    Examples:
        A frequency-axis normalization after a spectrogram:

        >>> model.add(Spectrogram())
        >>> model.add(Normalization2D(str_axis='freq'))

    """

    def __init__(
        self,
        str_axis: Optional[str] = None,
        int_axis: Optional[int] = None,
        image_data_format: str = "default",
        eps: float = 1e-10,
        **kwargs
    ):
        """[summary]

        Args:
            str_axis (Optional[str], optional):
                                    Axis name along which mean/std is computed (`batch`, `data_sample`, `channel`, `freq`, `time`).
                                    Recommended over `int_axis` because it provides more meaningful and image data format-robust interface.
                                    Defaults to None.
            int_axis (Optional[int], optional):
                                    Axis index along which mean/std is computed.
                                        - `0` for per data sample, `-1` for per batch.
                                        - `1`, `2`, `3` for channel, row, col (if channels_first)
                                    Defaults to None.
            image_data_format (str, optional): 'channels_first' (c,x,y,) or 'channels_last' (x,y,c) or TF 'default'. Defaults to 'default'.
            eps (float, optional): Small numerical value added to avoid divide by zero. Defaults to 1e-10.
        """
        assert not (int_axis is None and str_axis is None), "In Normalization2D, int_axis or str_axis should be specified."

        assert image_data_format in (
            "channels_first",
            "channels_last",
            "default",
        ), "Incorrect image_data_format: {}".format(image_data_format)

        if image_data_format == "default":
            self.image_data_format = K.image_data_format()
        else:
            self.image_data_format = image_data_format

        self.str_axis = str_axis
        if self.str_axis is None:  # use int_axis
            self.int_axis = int_axis
        else:  # use str_axis
            # warning
            if int_axis is not None:
                print("int_axis={} passed but is ignored, str_axis is used instead.".format(int_axis))
            # do the work
            assert str_axis in (
                "batch",
                "data_sample",
                "channel",
                "freq",
                "time",
            ), "Incorrect str_axis: {}".format(str_axis)
            if str_axis == "batch":
                int_axis = -1
            else:
                if self.image_data_format == "channels_first":
                    int_axis = ["data_sample", "channel", "freq", "time"].index(str_axis)
                else:
                    int_axis = ["data_sample", "freq", "time", "channel"].index(str_axis)

        assert int_axis in (-1, 0, 1, 2, 3), "invalid int_axis: " + str(int_axis)
        self.axis = int_axis
        self.eps = eps
        super(Normalization2D, self).__init__(**kwargs)

    def call(self, x, mask=None):
        if self.axis == -1:
            mean = K.mean(x, axis=[3, 2, 1, 0], keepdims=True)
            std = K.std(x, axis=[3, 2, 1, 0], keepdims=True)
        elif self.axis in (0, 1, 2, 3):
            all_dims = [0, 1, 2, 3]
            del all_dims[self.axis]
            mean = K.mean(x, axis=all_dims, keepdims=True)
            std = K.std(x, axis=all_dims, keepdims=True)
        return (x - mean) / (std + self.eps)

    def get_config(self):
        config = {
            "int_axis": self.axis,
            "str_axis": self.str_axis,
            "image_data_format": self.image_data_format,
        }
        base_config = super(Normalization2D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import logging
import tensorflow.keras as keras
import tensorflow as tf

try:
    import tensorflow_probability as tfp
except ImportError:
    tfp = None
    logging.debug("No tensorflow-probability.")
    pass


class MelSpec(keras.layers.Layer):
    # copied from https://keras.io/examples/audio/melgan_spectrogram_inversion

    def __init__(
        self,
        sampling_rate,
        frame_length=512,
        frame_step=None,
        fft_length=None,
        num_mel_channels=128,
        freq_min=0,
        freq_max=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.sampling_rate = sampling_rate
        self.frame_length = frame_length

        if frame_step is None:
            frame_step = self.frame_length // 4
        self.frame_step = frame_step

        if fft_length is None:
            fft_length = self.frame_length * 2
        self.fft_length = fft_length

        self.num_mel_channels = num_mel_channels
        self.freq_min = freq_min

        if freq_max is None:
            freq_max = self.sampling_rate // 2
        self.freq_max = freq_max

        # Defining mel filter. This filter will be multiplied with the STFT output
        self.mel_filterbank = tf.signal.linear_to_mel_weight_matrix(
            num_mel_bins=self.num_mel_channels,
            num_spectrogram_bins=self.fft_length // 2 + 1,
            sample_rate=self.sampling_rate,
            lower_edge_hertz=self.freq_min,
            upper_edge_hertz=self.freq_max,
        )

    def call(self, audio, training=True):
        # We will only perform the transformation during training.
        # if training:
        # Taking the Short Time Fourier Transform. Ensure that the audio is padded.
        # In the paper, the STFT output is padded using the 'REFLECT' strategy.
        stft = tf.signal.stft(
            tf.squeeze(audio, -1),
            self.frame_length,
            self.frame_step,
            self.fft_length,
            pad_end=True,
        )

        # Taking the magnitude of the STFT output
        magnitude = tf.abs(stft)

        # Multiplying the Mel-filterbank with the magnitude and scaling it using the db scale
        mel = tf.matmul(tf.square(magnitude), self.mel_filterbank)

        # # baseline subtract
        if tfp is not None:
            baseline = tfp.stats.percentile(mel, 50, axis=1, keepdims=True)
            mel = tf.math.log(mel / (baseline + 0.000001))
            mel = tf.clip_by_value(mel, clip_value_min=0, clip_value_max=100_000)
            mel /= tf.reduce_max(mel) * 255
        return mel

    def get_config(self):
        config = super(MelSpec, self).get_config()
        config.update(
            {
                "frame_length": self.frame_length,
                "frame_step": self.frame_step,
                "fft_length": self.fft_length,
                "sampling_rate": self.sampling_rate,
                "num_mel_channels": self.num_mel_channels,
                "freq_min": self.freq_min,
                "freq_max": self.freq_max,
            }
        )
        return config



import inspect
from typing import List

from tensorflow.keras import backend as K, Model, Input, optimizers
from tensorflow.keras import layers
from tensorflow.keras.layers import Activation, SpatialDropout1D, Lambda
from tensorflow.keras.layers import Layer, Conv1D, SeparableConv1D, Dense, BatchNormalization, LayerNormalization


def is_power_of_two(num: int):
    return num != 0 and ((num & (num - 1)) == 0)


def adjust_dilations(dilations: list):
    if all([is_power_of_two(i) for i in dilations]):
        return dilations
    else:
        new_dilations = [2 ** i for i in dilations]
        return new_dilations


def channel_normalization(x):
    # type: (Layer) -> Layer
    """Normalize a layer to the maximum activation

    This keeps a layers values between zero and one.
    It helps with relu's unbounded activation

    Args:
        x: The layer to normalize

    Returns:
        A maximal normalized layer
    """
    max_values = K.max(K.abs(x), 2, keepdims=True) + 1e-5
    out = x / max_values
    return out


class ResidualBlock(Layer):
    def __init__(
        self,
        dilation_rate: int,
        nb_filters: int,
        kernel_size: int,
        padding: str,
        activation: str = "relu",
        dropout_rate: float = 0,
        kernel_initializer: str = "he_normal",
        use_batch_norm: bool = False,
        use_layer_norm: bool = False,
        use_separable: bool = False,
        **kwargs
    ):
        """Defines the residual block for the WaveNet TCN

        Args:
            x: The previous layer in the model
            training: boolean indicating whether the layer should behave in training mode or in inference mode
            dilation_rate: The dilation power of 2 we are using for this residual block
            nb_filters: The number of convolutional filters to use in this block
            kernel_size: The size of the convolutional kernel
            padding: The padding used in the convolutional layers, 'same' or 'causal'.
            activation: The final activation used in o = Activation(x + F(x))
            dropout_rate: Float between 0 and 1. Fraction of the input units to drop.
            kernel_initializer: Initializer for the kernel weights matrix (Conv1D).
            use_batch_norm: Whether to use batch normalization in the residual layers or not.
            use_layer_norm: Whether to use layer normalization in the residual layers or not.
            use_separable: Whether to use separabel conv1d.
            kwargs: Any initializers for Layer class.
        """

        self.dilation_rate = dilation_rate
        self.nb_filters = nb_filters
        self.kernel_size = kernel_size
        self.padding = padding
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.use_batch_norm = use_batch_norm
        self.use_layer_norm = use_layer_norm
        self.use_separable = use_separable
        self.kernel_initializer = kernel_initializer
        self.layers = []
        self.layers_outputs = []
        self.shape_match_conv = None
        self.res_output_shape = None
        self.final_activation = None

        super(ResidualBlock, self).__init__(**kwargs)

    def _add_and_activate_layer(self, layer):
        """Helper function for building layer

        Args:
            layer: Appends layer to internal layer list and builds it based on the current output
                   shape of ResidualBlocK. Updates current output shape.

        """
        self.layers.append(layer)
        self.layers[-1].build(self.res_output_shape)
        self.res_output_shape = self.layers[-1].compute_output_shape(self.res_output_shape)

    def build(self, input_shape):

        with K.name_scope(self.name):  # name scope used to make sure weights get unique names
            self.layers = []
            self.res_output_shape = input_shape

            ConvThis = SeparableConv1D if self.use_separable else Conv1D

            for k in range(1):
                name = "conv1D_{}".format(k)
                with K.name_scope(name):  # name scope used to make sure weights get unique names
                    self._add_and_activate_layer(
                        ConvThis(
                            filters=self.nb_filters,
                            kernel_size=self.kernel_size,
                            dilation_rate=self.dilation_rate,
                            padding=self.padding,
                            name=name,
                            kernel_initializer=self.kernel_initializer,
                        )
                    )

                # with K.name_scope('norm_{}'.format(k)):
                #     if self.use_batch_norm:
                #         self._add_and_activate_layer(BatchNormalization())
                #     elif self.use_layer_norm:
                #         self._add_and_activate_layer(LayerNormalization())

                # self._add_and_activate_layer(Activation('relu'))
                # if self.activation == 'norm_relu':
                self._add_and_activate_layer(Activation("relu"))
                self._add_and_activate_layer(Lambda(channel_normalization))
                # else:
                #     self._add_and_activate_layer(Activation(self.activation))

                self._add_and_activate_layer(SpatialDropout1D(rate=self.dropout_rate))

            if self.nb_filters != input_shape[-1]:
                # 1x1 conv to match the shapes (channel dimension).
                name = "matching_conv1D"
                with K.name_scope(name):
                    # make and build this layer separately because it directly uses input_shape
                    self.shape_match_conv = Conv1D(
                        filters=self.nb_filters,
                        kernel_size=1,
                        padding="same",
                        name=name,
                        kernel_initializer=self.kernel_initializer,
                    )

            else:
                name = "matching_identity"
                self.shape_match_conv = Lambda(lambda x: x, name=name)

            with K.name_scope(name):
                self.shape_match_conv.build(input_shape)
                self.res_output_shape = self.shape_match_conv.compute_output_shape(input_shape)

            self.final_activation = Activation(self.activation)
            self.final_activation.build(self.res_output_shape)  # probably isn't necessary

            # this is done to force Keras to add the layers in the list to self._layers
            for layer in self.layers:
                self.__setattr__(layer.name, layer)
            self.__setattr__(self.shape_match_conv.name, self.shape_match_conv)
            self.__setattr__(self.final_activation.name, self.final_activation)

            super(ResidualBlock, self).build(input_shape)  # done to make sure self.built is set True

    def call(self, inputs, training=None):
        """
        Returns: A tuple where the first element is the residual model tensor, and the second
                 is the skip connection tensor.
        """
        x = inputs
        self.layers_outputs = [x]
        for layer in self.layers:
            training_flag = "training" in dict(inspect.signature(layer.call).parameters)
            x = layer(x, training=training) if training_flag else layer(x)
            self.layers_outputs.append(x)
        x2 = self.shape_match_conv(inputs)
        self.layers_outputs.append(x2)
        res_x = layers.add([x2, x])
        self.layers_outputs.append(res_x)

        res_act_x = self.final_activation(res_x)
        self.layers_outputs.append(res_act_x)
        return [res_act_x, x]

    def compute_output_shape(self, input_shape):
        return [self.res_output_shape, self.res_output_shape]


class TCN(Layer):
    """Creates a TCN layer.

    Input shape:
        A tensor of shape (batch_size, timesteps, input_dim).

    Args:
        nb_filters: The number of filters to use in the convolutional layers. Can be a list.
        kernel_size: The size of the kernel to use in each convolutional layer.
        dilations: The list of the dilations. Example is: [1, 2, 4, 8, 16, 32, 64].
        nb_stacks : The number of stacks of residual blocks to use.
        padding: The padding to use in the convolutional layers, 'causal' or 'same'.
        use_skip_connections: Boolean. If we want to add skip connections from input to each residual blocK.
        return_sequences: Boolean. Whether to return the last output in the output sequence, or the full sequence.
        activation: The activation used in the residual blocks o = Activation(x + F(x)).
        dropout_rate: Float between 0 and 1. Fraction of the input units to drop.
        kernel_initializer: Initializer for the kernel weights matrix (Conv1D).
        use_batch_norm: Whether to use batch normalization in the residual layers or not.
        use_separable: Whether to use separable convolution in the residual layers or not.
        kwargs: Any other arguments for configuring parent class Layer. For example "name=str", Name of the model.
                Use unique names when using multiple TCN.

    Returns:
        A TCN layer.
    """

    def __init__(
        self,
        nb_filters=64,
        kernel_size=2,
        nb_stacks=1,
        dilations=(1, 2, 4, 8, 16, 32),
        padding="causal",
        use_skip_connections=False,
        dropout_rate=0.0,
        return_sequences=False,
        activation="relu",
        kernel_initializer="he_normal",
        use_batch_norm=False,
        use_layer_norm=False,
        use_separable=False,
        **kwargs
    ):

        self.return_sequences = return_sequences
        self.dropout_rate = dropout_rate
        self.use_skip_connections = use_skip_connections
        self.dilations = dilations
        self.nb_stacks = nb_stacks
        self.kernel_size = kernel_size
        self.nb_filters = nb_filters
        self.activation = activation
        self.padding = padding
        self.kernel_initializer = kernel_initializer
        self.use_batch_norm = use_batch_norm
        self.use_layer_norm = use_layer_norm
        self.use_separable = use_separable
        self.skip_connections = []
        self.residual_blocks = []
        self.layers_outputs = []
        self.build_output_shape = None
        self.slicer_layer = None  # in case return_sequence=False
        self.output_slice_index = None  # in case return_sequence=False
        self.padding_same_and_time_dim_unknown = False  # edge case if padding='same' and time_dim = None

        if isinstance(self.nb_filters, list):
            assert len(self.nb_filters) == len(self.dilations)

        if padding != "causal" and padding != "same":
            raise ValueError("Only 'causal' or 'same' padding are compatible for this layer.")

        # initialize parent class
        super(TCN, self).__init__(**kwargs)

    @property
    def receptive_field(self):
        assert_msg = "The receptive field formula works only with power of two dilations."
        assert all([is_power_of_two(i) for i in self.dilations]), assert_msg
        return self.kernel_size * self.nb_stacks * self.dilations[-1]

    def build(self, input_shape):

        # member to hold current output shape of the layer for building purposes
        self.build_output_shape = input_shape

        # list to hold all the member ResidualBlocks
        self.residual_blocks = []
        total_num_blocks = self.nb_stacks * len(self.dilations)
        if not self.use_skip_connections:
            total_num_blocks += 1  # cheap way to do a false case for below

        for s in range(self.nb_stacks):
            for i, d in enumerate(self.dilations):
                res_block_filters = self.nb_filters[i] if isinstance(self.nb_filters, list) else self.nb_filters
                self.residual_blocks.append(
                    ResidualBlock(
                        dilation_rate=d,
                        nb_filters=res_block_filters,
                        kernel_size=self.kernel_size,
                        padding=self.padding,
                        activation=self.activation,
                        dropout_rate=self.dropout_rate,
                        use_batch_norm=self.use_batch_norm,
                        use_layer_norm=self.use_layer_norm,
                        use_separable=self.use_separable,
                        kernel_initializer=self.kernel_initializer,
                        name="residual_block_{}".format(len(self.residual_blocks)),
                    )
                )
                # build newest residual block
                self.residual_blocks[-1].build(self.build_output_shape)
                self.build_output_shape = self.residual_blocks[-1].res_output_shape

        # this is done to force keras to add the layers in the list to self._layers
        for layer in self.residual_blocks:
            self.__setattr__(layer.name, layer)

        self.output_slice_index = None
        if self.padding == "same":
            time = self.build_output_shape.as_list()[1]
            if time is not None:  # if time dimension is defined. e.g. shape = (bs, 500, input_dim).
                self.output_slice_index = int(self.build_output_shape.as_list()[1] / 2)
            else:
                # It will known at call time. c.f. self.call.
                self.padding_same_and_time_dim_unknown = True

        else:
            self.output_slice_index = -1  # causal case.
        self.slicer_layer = Lambda(lambda tt: tt[:, self.output_slice_index, :])

    def compute_output_shape(self, input_shape):
        """
        Overridden in case keras uses it somewhere... no idea. Just trying to avoid future errors.
        """
        if not self.built:
            self.build(input_shape)
        if not self.return_sequences:
            batch_size = self.build_output_shape[0]
            batch_size = batch_size.value if hasattr(batch_size, "value") else batch_size
            nb_filters = self.build_output_shape[-1]
            return [batch_size, nb_filters]
        else:
            # Compatibility tensorflow 1.x
            return [v.value if hasattr(v, "value") else v for v in self.build_output_shape]

    def call(self, inputs, training=None):
        x = inputs
        self.layers_outputs = [x]
        self.skip_connections = []
        for layer in self.residual_blocks:
            try:
                x, skip_out = layer(x, training=training)
            except TypeError:  # compatibility with tensorflow 1.x
                x, skip_out = layer(K.cast(x, "float32"), training=training)
            self.skip_connections.append(skip_out)
            self.layers_outputs.append(x)

        if self.use_skip_connections:
            x = layers.add(self.skip_connections)
            self.layers_outputs.append(x)

        # ????
        x = Activation("relu")(x)

        if not self.return_sequences:
            # case: time dimension is unknown. e.g. (bs, None, input_dim).
            if self.padding_same_and_time_dim_unknown:
                self.output_slice_index = K.shape(self.layers_outputs[-1])[1] // 2
            x = self.slicer_layer(x)
            self.layers_outputs.append(x)
        return x

    def get_config(self):
        """
        Returns the config of a the layer. This is used for saving and loading from a model
        :return: python dictionary with specs to rebuild layer
        """
        config = super(TCN, self).get_config()
        config["nb_filters"] = self.nb_filters
        config["kernel_size"] = self.kernel_size
        config["nb_stacks"] = self.nb_stacks
        config["dilations"] = self.dilations
        config["padding"] = self.padding
        config["use_skip_connections"] = self.use_skip_connections
        config["dropout_rate"] = self.dropout_rate
        config["return_sequences"] = self.return_sequences
        config["activation"] = self.activation
        config["use_batch_norm"] = self.use_batch_norm
        config["use_layer_norm"] = self.use_layer_norm
        config["kernel_initializer"] = self.kernel_initializer
        return config



from __future__ import absolute_import

__version__ = "0.1.4"
VERSION = __version__

from . import time_frequency
from . import backend
from . import backend_keras

from . import augmentation
from . import filterbank
from . import utils



"""Utilities for dealing with annotations."""
import numpy as np
import xarray as xr
import pandas as pd
from collections import UserDict
from typing import Optional, List, Dict, Any, Union


class Events(UserDict):
    """Utility class for dealing with annotations."""

    def __init__(
        self,
        data: Optional[Dict[str, List[float]]] = None,
        categories: Optional[Dict[str, str]] = None,
        add_names_from_categories: bool = True,
    ):
        # """Initializes Events class. Do not use. Use `from_df` for `from_lists` instead.

        # Args:
        #     data (Optional[Dict[str, List[float]]], optional): [description]. Defaults to None.
        #     categories (Optional[Dict[str, str]], optional): [description]. Defaults to None.
        #     add_names_from_categories (bool, optional): [description]. Defaults to True.
        # """
        if data is None:
            data = dict()

        super().__init__(data)

        for key, val in self.items():
            val = np.array(val)
            if val.ndim == 1:
                val = val[:, np.newaxis]
            if val.shape[1] == 1:
                val = np.concatenate((val, val), axis=1)

            self.data[key] = val
        self.categories = self._infer_categories()

        # drop nan
        self._drop_nan()

        # preserve cats from input
        if hasattr(data, "categories"):
            for name, cat in data.categories.items():
                if name in self:  # update only existing keys
                    self.categories[name] = cat

        # update cats from arg
        if categories is not None:
            for name, cat in categories.items():
                if name in self:  # update only existing keys
                    self.categories[name] = cat
                elif add_names_from_categories:
                    self.add_name(name=name, category=cat)

    @classmethod
    def from_df(cls, df, possible_event_names: Optional[List] = None):
        """[summary]

        Args:
            df (pd.DataFrame): with columns `name`, `start_seconds`, `end_seconds`.
            possible_event_names (list, optional): [description]. Defaults to [].

        Returns:
            Events: [description]
        """
        if possible_event_names is None:
            possible_event_names = []

        return cls.from_lists(
            df.name.values, df.start_seconds.values.astype(float), df.stop_seconds.values.astype(float), possible_event_names
        )

    @classmethod
    def from_lists(cls, names, start_seconds, stop_seconds, possible_event_names: Optional[List] = None):
        """[summary]

        Args:
            names ([type]): [description]
            start_seconds ([type]): [description]
            stop_seconds ([type]): [description]
            possible_event_names (list, optional): [description]. Defaults to [].

        Returns:
            [type]: [description]
        """
        if possible_event_names is None:
            possible_event_names = []
        unique_names = list(set(names))
        unique_names.extend(possible_event_names)
        dct = {name: [] for name in unique_names}

        for name, start_second, stop_second in zip(names, start_seconds, stop_seconds):
            dct[name].append([start_second, stop_second])

        return cls(dct)

    @classmethod
    def from_dataset(cls, ds):
        """[summary]

        Args:
            ds ([type]): [description]

        Returns:
            [type]: [description]
        """
        start_seconds = np.array(ds.event_times.sel(event_time="start_seconds").data)
        stop_seconds = np.array(ds.event_times.sel(event_time="stop_seconds").data)
        names = np.array(ds.event_names.data)
        if "possible_event_names" in ds.attrs:
            possible_event_names = ds.attrs["possible_event_names"]
        elif "possible_event_names" in ds.event_names.attrs:
            possible_event_names = ds.event_names.attrs["possible_event_names"]
        else:
            possible_event_names = []

        out = cls.from_lists(names, start_seconds, stop_seconds, possible_event_names)
        if "event_categories" in ds:
            cats = {str(cat.event_types.data): str(cat.event_categories.data) for cat in ds.event_categories}
            out = cls(out, categories=cats)
        return out

    @classmethod
    def from_predict(cls, events=None, segments=None):
        # to lists or df
        names = []
        start_seconds = []
        stop_seconds = []
        possible_names = []

        if (
            segments is not None
            and len(segments)
            and "sequence" in segments
            and len(segments["sequence"])
            and "names" in segments
        ):
            if type(segments["sequence"][0]) is not str and type(segments["sequence"][0]) is not np.str_:
                segment_names = [segments["names"][ii] for ii in segments["sequence"]]  # from ints to names
            else:
                segment_names = segments["sequence"]
            names.extend(segment_names)
            start_seconds.extend(segments["onsets_seconds"])
            stop_seconds.extend(segments["offsets_seconds"])
            possible_names.extend(segments["names"])

        if events is not None and len(events) and "sequence" in events and len(events["sequence"]) and "names" in events:
            if type(events["sequence"][0]) is not str and type(events["sequence"][0]) is not np.str_:
                event_names = [events["names"][ii] for ii in events["sequence"]]  # from ints to names
            else:
                event_names = events["sequence"]
            names.extend(event_names)
            start_seconds.extend(events["seconds"])
            stop_seconds.extend(events["seconds"])
            possible_names.extend(events["names"])

        # only keep unique
        possible_names = list(set(possible_names))

        out = cls.from_lists(names, start_seconds, stop_seconds, possible_names)
        return out

    def _init_df(self) -> pd.DataFrame:
        return pd.DataFrame(columns=["name", "start_seconds", "stop_seconds"])

    def _append_row(self, df, name, start_seconds, stop_seconds=None) -> pd.DataFrame:
        if stop_seconds is None:
            stop_seconds = start_seconds
        new_row = pd.DataFrame(np.array([name, start_seconds, stop_seconds])[np.newaxis, :], columns=df.columns)
        return pd.concat((df, new_row), ignore_index=True)

    def to_df(self, preserve_empty: bool = True) -> pd.DataFrame:
        """Convert to pandas.DataFrame

        Args:
            preserve_empty (bool, optional):
                In keeping with the convention that events have identical start and stop times and segments do not,
                empty events are coded with np.nan as both start and stop and
                empty segments are coded as np.nan as start and 0 as stop.
                `from_df()` will obey this convention - if both start and stop are np.nan,
                the name will be a segment,
                if only the start is np.nan (the stop does not matter), the name will be an event
                Defaults to True.

        Returns:
            pandas.DataFrame: with columns `name`, `start_seconds`, `stop_seconds`, one row per event.
        """
        df = self._init_df()
        for name in self.names:
            for start_second, stop_second in zip(self.start_seconds(name), self.stop_seconds(name)):
                df = self._append_row(df, name, start_second, stop_second)
        if preserve_empty:  # ensure we keep events without annotations
            for name, cat in zip(self.names, self.categories.values()):
                if name not in df.name.values:
                    stop_seconds = (
                        np.nan if cat == "event" else 0
                    )  # (np.nan, np.nan) -> empty events, (np.nan, some number) -> empty segments
                    df = self._append_row(df, name, start_seconds=np.nan, stop_seconds=stop_seconds)
        # make sure start and stop seconds are numeric
        df["start_seconds"] = pd.to_numeric(df["start_seconds"], errors="coerce")
        df["stop_seconds"] = pd.to_numeric(df["stop_seconds"], errors="coerce")
        return df

    def to_dataset(self) -> xr.Dataset:
        """Returns an xarray dataset.

        Returns:
            xr.Dataset: with the data arrays `event_names` and `event_times`.
        """
        df = self.to_df()
        names = df.name.values
        start_seconds = df.start_seconds.values.astype(float)
        stop_seconds = df.stop_seconds.values.astype(float)

        da_names = xr.DataArray(
            name="event_names",
            data=np.array(names, dtype="U128"),
            dims=[
                "index",
            ],
        )
        da_times = xr.DataArray(
            name="event_times",
            data=np.array([start_seconds, stop_seconds]).T,
            dims=["index", "event_time"],
            coords={"event_time": ["start_seconds", "stop_seconds"]},
        )

        ds = xr.Dataset({da.name: da for da in [da_names, da_times]})
        ds.attrs["time_units"] = "seconds"
        ds.attrs["possible_event_names"] = self.names  # ensure that we preserve even names w/o events that get lost in to_df
        return ds

    def add_name(
        self,
        name,
        category="segment",
        times=None,
        overwrite: bool = False,
        append: bool = False,
        sort_after_append: bool = False,
    ):
        """[summary]

        Args:
            name ([type]): [description]
            category (str, optional): [description]. Defaults to 'segment'.
            times ([type], optional): [description]. Defaults to None.
            overwrite (bool, optional): [description]. Defaults to False.
            append (bool, optional): [description]. Defaults to False.
            sort_after_append (bool, optional): [description]. Defaults to False.
        """
        if times is None:
            times = np.zeros((0, 2))

        if name not in self or (name in self and overwrite):
            self.update({name: times})
            self.categories[name] = category
        elif name in self and append:
            self[name] = np.append(self[name], times, axis=0)
            if sort_after_append:
                self[name].sort(axis=0)

    def delete_name(self, name):
        """[summary]

        Args:
            name ([type]): [description]
        """
        if name in self:
            del self[name]
        if name in self.categories:
            del self.categories[name]

    def add_time(self, name, start_seconds, stop_seconds=None):
        """[summary]

        Args:
            name ([type]): [description]
            start_seconds ([type]): [description]
            stop_seconds ([type], optional): [description]. Defaults to None.
        """
        if name not in self:
            raise ValueError(f"Song type {name} not added yet. Add via 'self.add_name(name, category)'")

        if stop_seconds is None:
            stop_seconds = start_seconds

        self[name] = np.insert(self[name], len(self[name]), sorted([start_seconds, stop_seconds]), axis=0)

    def move_time(self, name, old_time, new_time):
        """[summary]

        Args:
            name ([type]): [description]
            old_time ([type]): [description]
            new_time ([type]): [description]
        """
        self[name][self[name] == old_time] = new_time

    def delete_time(self, name, time, tol=0):
        nearest_start = self._find_nearest(self.start_seconds(name), time)
        index = np.where(self.start_seconds(name) == nearest_start)[0][0]

        if self.categories[name] == "segment":
            matching_stop = self.stop_seconds(name)[index]
            event_at_time = matching_stop > time
        elif self.categories[name] == "event":
            event_at_time = np.abs(time - nearest_start) < tol
        else:
            event_at_time = False

        if event_at_time:
            deleted_time = self[name][index, :]
            self[name] = np.delete(self[name], index, axis=0)
        else:
            deleted_time = None
        return deleted_time

    def select_range(self, name: str, t0: Optional[float] = None, t1: Optional[float] = None, strict: bool = True):
        """Get indices of events within the range.

        Need to start and stop after t0 and before t1 (non-inclusive bounds).

        Args:
            name (str): [description]
            t0 (float, optional): [description]
            t1 (float, optional): [description]
            strict (bool, optional): if true, only matches events that start AND stop within the range,
                           if false, matches events that start OR stop within the range

        Returns:
            List[uint]: List of indices of events within the range
        """

        if t0 is None:
            t0 = 0
        if t1 is None:
            t1 = np.inf

        if strict:
            within_range = np.logical_and(self.start_seconds(name) > t0, self.stop_seconds(name) < t1)
        else:
            starts_in_range = np.logical_and(self.start_seconds(name) > t0, self.start_seconds(name) < t1)
            stops_in_range = np.logical_and(self.stop_seconds(name) > t0, self.stop_seconds(name) < t1)
            within_range = np.logical_or(starts_in_range, stops_in_range)
        within_range_indices = np.where(within_range)[0]
        return within_range_indices

    def filter_range(self, name, t0, t1, strict: bool = False):
        """Returns events within the range.

        Need to start and stop after t0 and before t1 (non-inclusive bounds).

        Args:
            name ([type]): [description]
            t0 ([type]): [description]
            t1 ([type]): [description]
            strict (bool): if true, only matches events that start AND stop within the range,
                           if false, matches events that start OR stop within the range
        Returns:
            List[float]: [N, 2] list of start_seconds and stop_seconds in the range
        """
        indices = self.select_range(name, t0, t1, strict)
        return self[name][indices, :]

    def delete_range(self, name, t0, t1, strict: bool = True):
        """Deletes events within the range.

        Need to start and stop after t0 and before t1 (non-inclusive bounds).

        Args:
            name ([type]): [description]
            t0 ([type]): [description]
            t1 ([type]): [description]
            strict (bool): if true, only matches events that start AND stop within the range,
                           if false, matches events that start OR stop within the range
        Returns:
            int: number of deleted events
        """
        indices = self.select_range(name, t0, t1)
        self[name] = np.delete(self[name], indices, axis=0)
        return len(indices)

    def _find_nearest(self, array, value):
        array = np.asarray(array)
        idx = (np.abs(array - value)).argmin()
        return array[idx]

    def _infer_categories(self):
        categories = dict()
        for name in self.names:
            if len(self[name]) == 0:
                if not hasattr(self, "categories") or name not in self.categories:
                    categories[name] = None
                elif hasattr(self, "categories") and name in self.categories:
                    categories[name] = self.categories[name]
            else:
                first_start = self.start_seconds(name)[0]
                first_stop = self.stop_seconds(name)[0]

                if (np.isnan(first_start) and np.isnan(first_stop)) or (first_start == first_stop):
                    category = "event"
                else:
                    category = "segment"

                categories[name] = category

        return categories

    def _drop_nan(self):
        # remove entries with nan stop or start (but keep their name)
        for name in self.names:
            nan_events = np.logical_or(np.isnan(self.start_seconds(name)), np.isnan(self.stop_seconds(name)))
            self[name] = self[name][~nan_events]

    @property
    def names(self):
        return list(self.keys())

    def start_seconds(self, key):
        return self[key][:, 0]

    def stop_seconds(self, key):
        return self[key][:, 1]

    def duration_seconds(self, key):
        return self[key][:, 1] - self[key][:, 0]
